import time
from typing import Optional

import numpy as np

from .base_callback import BaseCallback


class ConvergenceTimeGetter(BaseCallback):
    """
    Monitor the training process and return the time it took to pass the threshold.
    """
    def __init__(
            self,
            *,
            metric: str,
            threshold: float,
            minimize_metric: bool,
            **kwargs
    ):
        """
        Constructor for ConvergenceTimeGetter class.

        :param metric: Name of the metric to monitor.
        :type metric: str
        :param threshold: Threshold value for the metric.
        :type threshold: float
        :param minimize_metric: Whether to minimize or maximize the metric.
        :type minimize_metric: bool
        :param kwargs: The keyword arguments to pass to the BaseCallback.
        """
        super().__init__(**kwargs)
        self.threshold = threshold
        self.metric = metric
        self.minimize_metric = minimize_metric
        self.threshold_met = False
        self.time_convergence = np.inf
        self.itr_convergence = np.inf
        self.training_time = np.inf
        self.start_time = None

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        if self.save_state:
            state = checkpoint.get(self.name, {})
            if state.get("threshold") == self.threshold and state.get("metric") == self.metric:
                super().load_checkpoint_state(trainer, checkpoint)
                self.start_time = time.time()
                if np.isfinite(state.get("time_convergence")):
                    self.start_time -= state.get("training_time", 0)
        # TODO: change start time and add training time, etc.

    def start(self, trainer, **kwargs):
        super().start(trainer)
        self.start_time = time.time()

    def close(self, trainer, **kwargs):
        self.training_time = time.time() - self.start_time

    def on_iteration_end(self, trainer, **kwargs):
        if not self.threshold_met:
            if self.minimize_metric:
                self.threshold_met = trainer.current_training_state.itr_metrics[self.metric] < self.threshold
            else:
                self.threshold_met = trainer.current_training_state.itr_metrics[self.metric] > self.threshold
            if self.threshold_met:
                self.save_on(trainer, **kwargs)

    def save_on(self, trainer, **kwargs):
        self.time_convergence = time.time() - self.start_time
        self.itr_convergence = trainer.current_training_state.iteration
        for cm in trainer.checkpoint_managers:
            cm.save_on(trainer)

    def __repr__(self):
        repr_str = f"ConvergenceTimeGetter("
        repr_str += f"metric={self.metric}, "
        repr_str += f"threshold={self.threshold}, "
        repr_str += f"minimize_metric={self.minimize_metric})"
        repr_str += f"<time_convergence={self.time_convergence} [s], "
        repr_str += f"itr_convergence={self.itr_convergence}, "
        repr_str += f"training_time={self.training_time} [s]>"
        return repr_str






import os.path
import pickle
import time
import multiprocessing as mp
from typing import Union, Optional

from matplotlib import pyplot as plt

from .base_callback import BaseCallback


class _VisualizerProcess(mp.Process):
    """
    Process to load a training history from pickle file and plot it. The process will update the plot every `update_dt`
    seconds and will close when join method is called.

    :Attributes:
        - **fig** (plt.Figure): The figure to plot.
        - **axes** (plt.Axes): The axes to plot.
        - **lines** (list): The list of lines to plot.
    """
    def __init__(self, history_path: str, lock: Union[mp.Lock, mp.RLock], update_dt: float = 1.0):
        """
        Create a new process to plot the training history.

        :param history_path: The path to the pickle file containing the training history.
        :type history_path: str
        :param lock: The lock to use to protect the training history.
        :type lock: Union[mp.Lock, mp.RLock]
        :param update_dt: The time in seconds between two updates of the plot.
        :type update_dt: float
        """
        super().__init__()
        self._history_path = history_path
        self._lock = lock
        self._update_dt = update_dt

        self._close_event = mp.Event()
        self._training_history = None
        self.fig, self.axes, self.lines = None, None, None

    def run(self):
        """
        Run the process.

        :return: None
        """
        can_plot = False
        while not can_plot and not self._close_event.is_set():
            if os.path.exists(self._history_path):
                try:
                    with self._lock:
                        self._training_history = pickle.load(open(self._history_path, "rb"))
                    can_plot = True
                except Exception:
                    time.sleep(self._update_dt)

        while len(self._training_history) < 2 and not self._close_event.is_set():
            self.update_history()
            time.sleep(self._update_dt)

        plt.ion()
        self.fig, self.axes, self.lines = self._training_history.create_plot()
        plt.pause(self._update_dt)
        while not self._close_event.is_set():
            self.update_history()
            self.fig, self.axes, self.lines = self._training_history.update_fig(self.fig, self.axes, self.lines)
            plt.pause(self._update_dt)
        plt.close(self.fig)

    def update_history(self):
        """
        Update the training history by loading it from the pickle file.

        :return: None
        """
        try:
            with self._lock:
                self._training_history = pickle.load(open(self._history_path, "rb"))
        except Exception as e:
            print(f"{e = }")
            pass

    def join(self, *args, **kwargs):
        """
        Set the close event.
        Join the current process.
        """
        self._close_event.set()
        plt.close(self.fig)
        return super().join(*args, **kwargs)


class TrainingHistoryVisualizationCallback(BaseCallback):
    """
    This callback is used to visualize the training history in real time.

    :Attributes:
        - **fig** (plt.Figure): The figure to plot.
        - **axes** (plt.Axes): The axes to plot.
        - **lines** (list): The list of lines to plot.
    """

    def __init__(
            self,
            temp_folder: str = '~/temp/',
            **kwargs
    ):
        """
        Create a new callback to visualize the training history.

        :param temp_folder: The folder where to save the training history.
        :type temp_folder: str
        :param kwargs: The keyword arguments to pass to the base callback.
        """
        super().__init__(**kwargs)
        self._is_open = False
        os.makedirs(temp_folder, exist_ok=True)
        self._temp_path = os.path.join(
            os.path.expanduser(temp_folder), f'training_visualization{time.time_ns()}.pkl'
        )
        self._lock = mp.Lock()
        self._process = _VisualizerProcess(self._temp_path, self._lock)

    def start(self, trainer, **kwargs):
        """
        Start the process.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        self._process.start()
        self._is_open = True

    def on_iteration_end(self, trainer, **kwargs):
        """
        Update the training history.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        with self._lock:
            pickle.dump(trainer.training_history, open(self._temp_path, "wb"))

    def __del__(self):
        """
        Delete the temporary file.

        :return: None
        """
        self.close(None)

    def close(self, trainer, **kwargs):
        """
        Close the process adn delete the temporary file.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        if self._is_open:
            self._process.join()
            self._process.kill()
            self._process.close()
            self._is_open = False
            with self._lock:
                if os.path.exists(self._temp_path):
                    os.remove(self._temp_path)






import enum
import json
import logging
import os
import pprint
import shutil
import warnings
from typing import Any, Dict, Optional, Union

import numpy as np
import torch

from .base_callback import BaseCallback
from ..utils import mapping_update_recursively


class LoadCheckpointMode(enum.Enum):
    """
    Enum for the different modes of loading a checkpoint.

    :Attributes:
        - **BEST_ITR** (int): Indicates that the iteration with the best metric will be loaded.
        - **LAST_ITR** (int): Indicates that the last iteration will be loaded.
    """
    BEST_ITR = 0
    LAST_ITR = 1

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(mode_name: str) -> 'LoadCheckpointMode':
        """
        Converts a string to a :class:`LoadCheckpointMode` instance.

        :param mode_name: The name of the mode.
        :type mode_name: str
        :return: The corresponding :class:`LoadCheckpointMode` instance.
        :rtype: LoadCheckpointMode
        """
        if mode_name.lower() in ["best", "last"]:
            mode_name = mode_name.upper() + "_ITR"
        assert mode_name.upper() in LoadCheckpointMode.__members__, f"Invalid mode name: {mode_name}"
        return LoadCheckpointMode[mode_name.upper()]


class CheckpointManager(BaseCallback):
    """
    This class is used to manage and create the checkpoints of a model.

    :Attributes:
        - **checkpoint_folder** (str): The folder to save the checkpoints to.
        - **meta_path_prefix** (str): The prefix to use for the checkpoint's metadata file.
        - **metric** (str): The name of the metric to collect the best checkpoint on.
        - **minimise_metric** (bool): Whether to minimise the metric or maximise it.
        - **curr_best_metric** (float): The current best metric value.
    """

    DEFAULT_PRIORITY = BaseCallback.DEFAULT_LOW_PRIORITY

    SAVE_EXT: str = '.pth'
    SUFFIX_SEP: str = '-'
    CHECKPOINTS_META_SUFFIX: str = 'checkpoints'
    CHECKPOINT_SAVE_PATH_KEY: str = "save_path"
    CHECKPOINT_BEST_KEY: str = "best"
    CHECKPOINT_ITRS_KEY: str = "iterations"
    CHECKPOINT_ITR_KEY: str = "itr"
    CHECKPOINT_METRICS_KEY: str = 'metrics'
    CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
    CHECKPOINT_STATE_DICT_KEY: str = "model_state_dict"
    CHECKPOINT_TRAINING_HISTORY_KEY: str = "training_history"
    CHECKPOINT_FILE_STRUCT: Dict[str, Union[str, Dict[int, str]]] = {
        CHECKPOINT_BEST_KEY: CHECKPOINT_SAVE_PATH_KEY,
        CHECKPOINT_ITRS_KEY: {0: CHECKPOINT_SAVE_PATH_KEY},
    }
    load_mode_to_suffix: Dict[LoadCheckpointMode, str] = {mode: mode.name for mode in list(LoadCheckpointMode)}

    @staticmethod
    def _replace_trainer_history(trainer, new_history: Any):
        """
        Replaces the training history in the trainer with the given history.
        :param trainer: The trainer object.
        :param new_history: The new history object.
        :return: None
        """
        # trainer.callbacks.remove(trainer.training_history)
        # trainer.callbacks.append(new_history)
        # trainer.training_history = new_history
        trainer.training_history.load_checkpoint_state(trainer, new_history)
        trainer.sort_callbacks_()

    @staticmethod
    def get_save_name_from_checkpoints(
            checkpoints_meta: Dict[str, Union[str, Dict[Any, str]]],
            load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR
    ) -> str:
        """
        Gets the save name from the checkpoint's metadata given the load checkpoint mode.

        :param checkpoints_meta: The checkpoint's metadata.
        :type checkpoints_meta: Dict[str, Union[str, Dict[Any, str]]]
        :param load_checkpoint_mode: The load checkpoint mode.
        :type load_checkpoint_mode: LoadCheckpointMode

        :return: The save name.
        :rtype: str
        """
        if load_checkpoint_mode == load_checkpoint_mode.BEST_ITR:
            if CheckpointManager.CHECKPOINT_BEST_KEY in checkpoints_meta:
                return checkpoints_meta[CheckpointManager.CHECKPOINT_BEST_KEY]
            else:
                raise FileNotFoundError(
                    f"No best checkpoint found in checkpoints_meta. "
                    f"Please use a different load_checkpoint_mode."
                )
        elif load_checkpoint_mode == load_checkpoint_mode.LAST_ITR:
            itr_dict = checkpoints_meta[CheckpointManager.CHECKPOINT_ITRS_KEY]
            last_itr: int = max([int(e) for e in itr_dict])
            return checkpoints_meta[CheckpointManager.CHECKPOINT_ITRS_KEY][str(last_itr)]
        else:
            raise ValueError("Invalid load_checkpoint_mode")

    def __init__(
            self,
            checkpoint_folder: str = "./checkpoints",
            *,
            checkpoints_meta_path: Optional[str] = None,
            meta_path_prefix: str = "network",
            metric: str = "val_loss",
            minimise_metric: bool = True,
            save_freq: int = 1,
            save_best_only: bool = False,
            start_save_at: int = 0,
            verbose: bool = False,
            **kwargs
    ):
        """
        Initialises the checkpoint manager.

        :param checkpoint_folder: The folder to save the checkpoints to.
        :type checkpoint_folder: str
        :param checkpoints_meta_path: The path to the checkpoints metadata file. If None, will use the
                                        `checkpoint_folder` and `meta_path_prefix` to create the path.
        :type checkpoints_meta_path: Optional[str]
        :param meta_path_prefix: The prefix to use for the checkpoint's metadata file.
        :type meta_path_prefix: str
        :param metric: The name of the metric to collect the best checkpoint on.
        :type metric: str
        :param minimise_metric: Whether to minimise the metric or maximise it.
        :type minimise_metric: bool
        :param save_freq: The frequency at which to save checkpoints. If set to <= 0, will save at the end of the
                            training.
        :type save_freq: int
        :param save_best_only: Whether to only save the best checkpoint.
        :type save_best_only: bool
        :param start_save_at: The iteration at which to start saving checkpoints.
        :type start_save_at: int
        :param verbose: Whether to print out the trace of the checkpoint manager.
        :type verbose: bool
        :param kwargs: The keyword arguments to pass to the BaseCallback.

        :keyword: **show_best_metric_on_p_bar** (bool): Whether to show the best metric on the progress bar.
            Default: `save_best_only`.
        """
        kwargs.setdefault("save_state", False)
        super().__init__(**kwargs)
        os.makedirs(checkpoint_folder, exist_ok=True)
        self.checkpoint_folder = checkpoint_folder
        self._checkpoints_meta_path = checkpoints_meta_path
        self.meta_path_prefix = meta_path_prefix
        self.verbose = verbose

        self.metric = metric
        self.minimise_metric = minimise_metric
        self.save_freq = save_freq
        self.save_best_only = save_best_only
        if self.save_best_only:
            self.save_freq = -1
        self.start_save_at = start_save_at
        self.curr_best_metric = np.inf if self.minimise_metric else -np.inf
        self.curr_checkpoint = None
        self.show_best_metric_on_p_bar = kwargs.get("show_best_metric_on_p_bar", self.save_best_only)

    @property
    def checkpoints_meta_path(self) -> str:
        """
        Gets the path to the checkpoints metadata file.

        :return: The path to the checkpoints metadata file.
        :rtype: str
        """
        if self._checkpoints_meta_path is not None:
            return self._checkpoints_meta_path
        full_filename = (
            f"{self.meta_path_prefix}{CheckpointManager.SUFFIX_SEP}{CheckpointManager.CHECKPOINTS_META_SUFFIX}"
        )
        return f"{self.checkpoint_folder}/{full_filename}.json"

    @checkpoints_meta_path.setter
    def checkpoints_meta_path(self, value: str):
        """
        Sets the path to the checkpoints metadata file.

        :param value: The path to the checkpoints metadata file.
        :type value: str

        :return: None
        """
        self._checkpoints_meta_path = value

    def get_checkpoint_filename(self, itr: int = -1):
        """
        Generate the filename for the checkpoint at the given iteration.

        :param itr: The iteration to generate the filename for.
        :type itr: int

        :return: The filename for the checkpoint at the given iteration.
        :rtype: str
        """
        pre_name = f"{self.meta_path_prefix}"
        if itr == -1:
            post_name = ""
        else:
            post_name = f"{CheckpointManager.SUFFIX_SEP}{CheckpointManager.CHECKPOINT_ITR_KEY}{itr}"
        return f"{pre_name}{post_name}{CheckpointManager.SAVE_EXT}"

    def _create_new_checkpoint_meta(self, itr: int, best: bool = False) -> dict:
        """
        Creates a new checkpoint's metadata.

        :param itr: The iteration of the checkpoint.
        :type itr: int
        :param best: Whether the checkpoint is currently the best.
        :type best: bool

        :return: The new checkpoint's metadata.
        :rtype: dict
        """
        save_name = self.get_checkpoint_filename(itr)
        new_info = {CheckpointManager.CHECKPOINT_ITRS_KEY: {str(itr): save_name}}
        if best:
            new_info[CheckpointManager.CHECKPOINT_BEST_KEY] = save_name
        return new_info

    def save_checkpoint(
            self,
            itr: int,
            itr_metrics: Dict[str, Any],
            best: bool = False,
            state_dict: Optional[Dict[str, Any]] = None,
            optimizer_state_dict: Optional[Dict[str, Any]] = None,
            training_history: Optional[Any] = None,
            **other_states,
    ) -> str:
        """
        Saves a checkpoint of the model and optimizer states at the given iteration.

        :param itr: The iteration number.
        :type itr: int
        :param itr_metrics: The metrics at the given iteration.
        :type itr_metrics: Dict[str, Any]
        :param best: Whether this is the best iteration so far.
        :type best: bool
        :param state_dict: The state dict of the model.
        :type state_dict: Optional[Dict[str, Any]]
        :param optimizer_state_dict: The state dict of the optimizer.
        :type optimizer_state_dict: Optional[Dict[str, Any]]
        :param training_history: The training history object.
        :type training_history: Optional[Any]

        :return: The path to the saved checkpoint.
        :rtype: str
        """
        os.makedirs(self.checkpoint_folder, exist_ok=True)
        save_name = self.get_checkpoint_filename(itr)
        path = os.path.join(self.checkpoint_folder, save_name)
        basic_states = {
            CheckpointManager.CHECKPOINT_ITR_KEY: itr,
            CheckpointManager.CHECKPOINT_STATE_DICT_KEY: state_dict,
            CheckpointManager.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: optimizer_state_dict,
            CheckpointManager.CHECKPOINT_METRICS_KEY: itr_metrics,
            CheckpointManager.CHECKPOINT_TRAINING_HISTORY_KEY: training_history,
        }
        assert all(key not in other_states for key in basic_states.keys()), (
            f"Other states cannot have the same keys as the basic states ({basic_states.keys()})."
        )
        states = {**basic_states, **other_states}
        torch.save(states, path)
        self.save_checkpoints_meta(self._create_new_checkpoint_meta(itr, best))
        return path

    def load_checkpoint(
            self,
            load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR
    ) -> dict:
        """
        Loads the checkpoint at the given load_checkpoint_mode.

        :param load_checkpoint_mode: The load_checkpoint_mode to use.
        :type load_checkpoint_mode: LoadCheckpointMode

        :return: The loaded checkpoint.
        :rtype: dict
        """
        # TODO: add the possibility to load a specific itr
        with open(self.checkpoints_meta_path, "r+") as jsonFile:
            info: dict = json.load(jsonFile)
        filename = CheckpointManager.get_save_name_from_checkpoints(info, load_checkpoint_mode)
        checkpoint = torch.load(f"{self.checkpoint_folder}/{filename}")
        return checkpoint

    def save_checkpoints_meta(self, new_info: dict):
        """
        Saves the new checkpoints' metadata.

        :param new_info: The new checkpoints' metadata.
        :type new_info: dict

        :return: None
        """
        info = dict()
        if os.path.exists(self.checkpoints_meta_path):
            with open(self.checkpoints_meta_path, "r+") as jsonFile:
                info = json.load(jsonFile)
        mapping_update_recursively(info, new_info)
        os.makedirs(self.checkpoint_folder, exist_ok=True)
        with open(self.checkpoints_meta_path, "w+") as jsonFile:
            json.dump(info, jsonFile, indent=4)

    def start(self, trainer, **kwargs):
        """
        Call at the beginning of the training by the Trainer. Load the checkpoint base on the load_checkpoint_mode of
        the trainer and update the current_training_state of the trainer.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        super().start(trainer)
        if trainer.load_checkpoint_mode is None:
            trainer.load_checkpoint_mode = LoadCheckpointMode.LAST_ITR
        start_itr = 0
        checkpoint = self.curr_checkpoint
        if trainer.force_overwrite:
            if os.path.exists(self.checkpoints_meta_path):
                shutil.rmtree(self.checkpoint_folder)
        else:
            try:
                checkpoint = self.load_checkpoint(trainer.load_checkpoint_mode)
                trainer.model.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_STATE_DICT_KEY], strict=True)
                if trainer.optimizer is not None:
                    trainer.optimizer.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY])
                start_itr = int(checkpoint[CheckpointManager.CHECKPOINT_ITR_KEY]) + 1
            # self._replace_trainer_history(trainer, checkpoint[CheckpointManager.CHECKPOINT_TRAINING_HISTORY_KEY])
            except FileNotFoundError as e:
                if self.verbose:
                    logging.info("No such checkpoint. Fit from beginning.")
            finally:
                self.curr_checkpoint = checkpoint

        trainer.update_state_(iteration=start_itr)
        if self.minimise_metric:
            self.curr_best_metric = trainer.training_history.min(self.metric)
        else:
            self.curr_best_metric = trainer.training_history.max(self.metric)

    def save_on(self, trainer) -> bool:
        """
        Saves the checkpoint if the current iteration is a checkpoint iteration.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: Whether the checkpoint was saved.
        :rtype: bool
        """
        itr_metrics = trainer.current_training_state.itr_metrics
        if itr_metrics is None:
            itr_metrics = {}
        other_states = trainer.callbacks.get_checkpoint_state(trainer)
        is_best = self._check_is_best(trainer)
        if is_best:
            self.curr_best_metric = itr_metrics[self.metric]
        self.save_checkpoint(
            trainer.current_training_state.iteration, itr_metrics, is_best,
            state_dict=trainer.model.state_dict(),
            optimizer_state_dict=trainer.optimizer.state_dict() if trainer.optimizer else None,
            training_history=trainer.training_history.get_checkpoint_state(trainer),
            **other_states
        )
        if trainer.training_history:
            trainer.training_history.plot(
                save_path=os.path.join(self.checkpoint_folder, "training_history.png"),
                show=False
            )
        return True

    def on_iteration_end(self, trainer, **kwargs):
        """
        Called when an iteration ends. An iteration is defined as one full pass through the training dataset and
        the validation dataset. The checkpoint is saved if the current constraints are met.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        if self.start_save_at > trainer.current_training_state.iteration:
            return

        if self.save_best_only:
            if self._check_is_best(trainer):
                self.save_on(trainer)
            return

        if self.save_freq > 0 and trainer.current_training_state.iteration % self.save_freq == 0:
            return self.save_on(trainer)
        if trainer.current_training_state.iteration >= trainer.state.n_iterations - 1:
            return self.save_on(trainer)

    def _check_is_best(self, trainer) -> Optional[bool]:
        if trainer.current_training_state.itr_metrics is None:
            return None

        itr_metric = trainer.current_training_state.itr_metrics.get(self.metric, self.curr_best_metric)
        if self.minimise_metric:
            is_best = itr_metric < self.curr_best_metric
        else:
            is_best = itr_metric > self.curr_best_metric
        return is_best

    def close(self, trainer, **kwargs):
        """
        Called when the training is finished. Saves the current checkpoint if the current iteration is lower than
        the number of iterations i.e. there is new stuff to save.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        if trainer.current_training_state.iteration < trainer.state.n_iterations:
            self.save_on(trainer)

    def extra_repr(self) -> str:
        extra_repr = f"folder={self.checkpoint_folder}, metric={self.metric}, minimise={self.minimise_metric}"
        extra_repr += f", start_save_at={self.start_save_at}"
        extra_repr += f", save_freq={self.save_freq}"
        extra_repr += f", save_best_only={self.save_best_only}"
        return extra_repr

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        """
        Called when the progress bar is updated. Adds the current best metric to the progress bar.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: The dictionary to add to the progress bar.
        :rtype: dict
        """
        if self.show_best_metric_on_p_bar:
            return {f"best_{self.metric}": self.curr_best_metric}
        return {}



import warnings
from collections import OrderedDict
from copy import deepcopy, copy
from typing import Iterable, Optional, List, Callable, Dict, Any, Union, NamedTuple, Generator

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from ..transforms.base import to_numpy, ToTensor
from ..learning_algorithms.bptt import BPTT
from ..callbacks import CheckpointManager, LoadCheckpointMode, TrainingHistory
from ..callbacks.base_callback import BaseCallback, CallbacksList
from ..learning_algorithms.learning_algorithm import LearningAlgorithm
from ..modules import BaseModel
from ..regularization import BaseRegularization, RegularizationList
from ..utils.collections import unpack_x_hh_y


class CurrentTrainingState(NamedTuple):
    r"""
    This class is used to store the current training state. It is extremely useful for the callbacks
    to access the current training state and to personalize the training process.

    :Attributes:
        - **n_iterations** (int): The total number of iterations.
        - **iteration** (int): The current iteration.
        - **epoch** (int): The current epoch.
        - **batch** (int): The current batch.
        - **x_batch** (Any): The current input batch.
        - **hh_batch** (Any): The current hidden state batch.
        - **y_batch** (Any): The current target batch.
        - **pred_batch** (Any): The current prediction.
        - **batch_loss** (float): The current loss.
        - **batch_is_train** (bool): Whether the current batch is a training batch.
        - **train_loss** (float): The current training loss.
        - **val_loss** (float): The current validation loss.
        - **itr_metrics** (Dict[str, Any]): The current iteration metrics.
        - **stop_training_flag** (bool): Whether the training should be stopped.
        - **info** (Dict[str, Any]): Any additional information. This is useful to communicate between callbacks.
        - **objects** (Dict[str, Any]): Any additional objects. This is useful to manage objects between callbacks.
            Note: In general, the train_dataloader and val_dataloader should be stored here.

    """
    n_iterations: Optional[int] = None
    iteration: Optional[int] = None
    n_epochs: Optional[int] = None
    epoch: Optional[int] = None
    epoch_loss: Optional[Any] = None
    batch: Optional[int] = None
    x_batch: Optional[Any] = None
    hh_batch: Optional[Any] = None
    y_batch: Optional[Any] = None
    pred_batch: Optional[Any] = None
    batch_loss: Optional[Any] = None
    batch_is_train: Optional[bool] = None
    train_loss: Optional[Any] = None
    val_loss: Optional[Any] = None
    train_metrics: Optional[Any] = None
    val_metrics: Optional[Any] = None
    itr_metrics: Optional[Dict[str, Any]] = {}
    stop_training_flag: bool = False
    info: Dict[str, Any] = {}
    objects: Dict[str, Any] = {}

    def __getstate__(self):
        not_picklable = ["info", "objects"]
        d = {k: v for k, v in self._asdict().items() if k not in not_picklable}
        return d

    @staticmethod
    def get_null_state() -> "CurrentTrainingState":
        return CurrentTrainingState()

    def update(self, **kwargs) -> "CurrentTrainingState":
        self_dict = self._asdict()
        assert all(k in self_dict for k in kwargs)
        self_dict.update(kwargs)
        return CurrentTrainingState(**self_dict)


TrainingState = CurrentTrainingState  # Alias


class Trainer:
    """
    Trainer class. This class is used to train a model.

    TODO: Add the possibility to pass a callable as the `predict_method`.
    """
    def __init__(
            self,
            model: torch.nn.Module,
            *,
            predict_method: str = "__call__",
            criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
            regularization: Optional[Union[BaseRegularization, RegularizationList, Iterable[BaseRegularization]]] = None,
            optimizer: Optional[torch.optim.Optimizer] = None,
            learning_algorithm: Optional[LearningAlgorithm] = None,
            regularization_optimizer: Optional[torch.optim.Optimizer] = None,
            metrics: Optional[List[Callable]] = None,
            callbacks: Optional[Union[List[BaseCallback], CallbacksList, BaseCallback]] = None,
            device: Optional[torch.device] = None,
            verbose: bool = True,
            **kwargs
    ):
        """
        Constructor for Trainer.

        :param model: Model to train.
        :param criterion: Loss function(s) to use. Deprecated, use `learning_algorithm` instead.
        :param regularization: Regularization(s) to use. In NeuroTorch, there are two ways to do regularization:
            1. Regularization can be specified in the layers with the 'update_regularization_loss' method. This
            regularization will be performed by the same optimizer as the main loss. This way is useful when you
            want a regularization that depends on the model output or hidden state.
            2. Regularization can be specified in the trainer with the 'regularization' parameter. This regularization
            will be performed by a separate optimizer named 'regularization_optimizer'. This way is useful when you
            want a regularization that depends only on the model parameters and when you want to control the
            learning rate of the regularization independently of the main loss.

            Note: This parameter will be deprecated and remove in a future version. The regularization will be
                specified in the learning algorithm and/or in the callbacks.

        :param optimizer: Optimizer to use for the main loss. Deprecated. Use learning_algorithm instead.
        :param learning_algorithm: Learning algorithm to use for the main loss. This learning algorithm can be given
            in the callbacks list as well. If specified, this learning algorithm will be added to the callbacks list.
            In this case, make sure that the learning algorithm is not added twice. Note that multiple learning
            algorithms can be used in the callbacks list.
        :param regularization_optimizer: Optimizer to use for the regularization loss.
        :param metrics: Metrics to compute during training.
        :param callbacks: Callbacks to use during training. Each callback will be called at different moments,
            see the documentation of :class:`BaseCallback` for more information.
        :param device: Device to use for the training. Default is the device of the model.
        :param verbose: Whether to print information during training.
        :param kwargs: Additional arguments of the training.

        :keyword int n_epochs: The number of epochs to train at each iteration. Default is 1.
        :keyword float lr: Learning rate of the main optimizer. Default is 1e-3.
        :keyword float reg_lr: Learning rate of the regularization optimizer. Default is 1e-2.
        :keyword float weight_decay: Weight decay of the main optimizer. Default is 0.0.
        :keyword bool exec_metrics_on_train: Whether to compute metrics on the train dataset. This is useful when
            you want to save time by not computing the metrics on the train dataset. Default is True.
        :keyword x_transform: Transform to apply to the input data before passing it to the model.
        :keyword y_transform: Transform to apply to the target data before passing it to the model. For example,
            this can be used to convert the target data to a one-hot encoding or to long tensor
            using `nt.ToTensor(dtype=torch.long)`.
        """
        # assert model.is_built, "Model must be built before training"
        self.kwargs = self._set_default_kwargs(kwargs)
        self.model = model
        self.predict_method = predict_method
        assert hasattr(model, predict_method), f"Model {model.__class__} does not have a method named '{predict_method}'"
        assert callable(getattr(model, predict_method)), f"Model method '{model.__class__}.{predict_method}' is not callable"
        self.criterion = self._set_default_criterion(criterion)
        self.regularization = self._set_default_regularization(regularization)
        # self._maybe_add_regularization(self.regularization)
        self.optimizer = self._set_default_optimizer(optimizer)
        self.regularization_optimizer = self._set_default_reg_optimizer(regularization_optimizer)
        self.metrics = self._set_default_metrics(metrics)
        self.callbacks: CallbacksList = self._set_default_callbacks(callbacks)
        self._maybe_add_learning_algorithm(learning_algorithm)
        self.sort_callbacks_()
        self.device = self._set_default_device(device)
        self.verbose = verbose
        self.training_history: TrainingHistory = self.training_histories[0]
        self.current_training_state = CurrentTrainingState()

        self.x_transform = kwargs.get("x_transform", ToTensor())
        self.y_transform = kwargs.get("y_transform", ToTensor())

        self._load_checkpoint_mode = None
        self._force_overwrite = None

    @property
    def network(self):
        """
        Alias for the model.

        :return: The :attr:`model` attribute.
        """
        return self.model

    @network.setter
    def network(self, value):
        """
        Alias for the model.

        :param value: The new value for the :attr:`model` attribute.
        :return: None
        """
        self.model = value

    @property
    def state(self):
        """
        Alias for the :attr:`current_training_state` attribute.

        :return: The :attr:`current_training_state`
        """
        return self.current_training_state

    @property
    def load_checkpoint_mode(self):
        return self._load_checkpoint_mode

    @load_checkpoint_mode.setter
    def load_checkpoint_mode(self, value: LoadCheckpointMode):
        self._load_checkpoint_mode = value

    @property
    def force_overwrite(self):
        return self._force_overwrite

    @property
    def training_histories(self) -> CallbacksList:
        return CallbacksList(list(filter(lambda x: isinstance(x, TrainingHistory), self.callbacks)))

    @property
    def checkpoint_managers(self) -> CallbacksList:
        return CallbacksList(list(filter(lambda x: isinstance(x, CheckpointManager), self.callbacks)))

    @property
    def learning_algorithms(self) -> CallbacksList:
        return CallbacksList(list(filter(lambda x: isinstance(x, LearningAlgorithm), self.callbacks)))

    @staticmethod
    def _set_default_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        kwargs.setdefault("n_epochs", 1)
        kwargs.setdefault("lr", 1e-3)
        kwargs.setdefault("reg_lr", 1e-2)
        kwargs.setdefault("weight_decay", 0.0)
        kwargs.setdefault("batch_size", 256)
        kwargs.setdefault("exec_metrics_on_train", True)

        assert kwargs["batch_size"] > 0, "batch_size must be positive"
        return kwargs

    def _set_default_optimizer(self, optimizer: Optional[torch.optim.Optimizer]) -> torch.optim.Optimizer:
        warnings.warn("The 'optimizer' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
        # if optimizer is None:
        # 	optimizer = torch.optim.Adam(
        # 		self.model.parameters(),
        # 		lr=self.kwargs["lr"],
        # 		weight_decay=self.kwargs["weight_decay"],
        # 	)
        return optimizer

    def _maybe_add_learning_algorithm(self, learning_algorithm: Optional[LearningAlgorithm]) -> None:
        if len(self.learning_algorithms) == 0 and learning_algorithm is None:
            learning_algorithm = BPTT(optimizer=self.optimizer, criterion=self.criterion)
        if learning_algorithm is not None:
            self.callbacks.append(learning_algorithm)

    def _maybe_add_regularization(self, regularization: Optional[RegularizationList]) -> None:
        if regularization is not None:
            self.callbacks.append(regularization)

    def _set_default_reg_optimizer(self, optimizer: Optional[torch.optim.Optimizer]) -> torch.optim.Optimizer:
        warnings.warn("The 'regularization_optimizer' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
        if optimizer is None and self.regularization is not None:
            optimizer = torch.optim.SGD(
                self.regularization.parameters(),
                lr=self.kwargs["reg_lr"],
                weight_decay=0.0,
            )
        return optimizer

    def _set_default_metrics(self, metrics: Optional[List[Callable]]):
        if metrics is None:
            metrics = []
        return metrics

    def _set_default_criterion(self, criterion: Optional[torch.nn.Module]) -> torch.nn.Module:
        warnings.warn("The 'criterion' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
        # if criterion is None:
        # 	if isinstance(self.model.output_sizes, dict):
        # 		criterion = {
        # 			k: torch.nn.MSELoss() for k in self.model.output_sizes
        # 		}
        # 	elif isinstance(self.model.output_sizes, int):
        # 		criterion = torch.nn.MSELoss()
        # 	else:
        # 		raise ValueError("Unknown criterion type")
        return criterion

    def _set_default_regularization(
            self,
            regularization: Optional[Union[BaseRegularization, RegularizationList, Iterable[BaseRegularization]]]
    ) -> Optional[RegularizationList]:
        warnings.warn("The 'regularization' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
        if regularization is None:
            pass
        elif isinstance(regularization, BaseRegularization):
            regularization = RegularizationList([regularization])
        elif isinstance(regularization, RegularizationList):
            pass
        elif isinstance(regularization, Iterable):
            regularization = RegularizationList(regularization)
        return regularization

    def _set_default_device(self, device: Optional[torch.device]) -> torch.device:
        if device is None:
            if hasattr(self.model, "device"):
                device = self.model.device
            else:
                device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        return device

    @staticmethod
    def _set_default_callbacks(
            callbacks: Optional[Union[List[BaseCallback], CallbacksList, BaseCallback]]
    ) -> CallbacksList:
        if callbacks is None:
            callbacks = []
        if isinstance(callbacks, BaseCallback):
            callbacks = [callbacks]
        if not any([isinstance(callback, TrainingHistory) for callback in callbacks]):
            callbacks.append(TrainingHistory())
        return CallbacksList(callbacks)

    def update_state_(self, **kwargs):
        self.current_training_state = self.current_training_state.update(**kwargs)

    def update_objects_state_(self, **kwargs):
        self.update_state_(objects={**self.current_training_state.objects, **kwargs})

    def update_info_state_(self, **kwargs):
        self.update_state_(info={**self.current_training_state.info, **kwargs})

    def update_itr_metrics_state_(self, **kwargs):
        self.update_state_(itr_metrics={**self.current_training_state.itr_metrics, **kwargs})

    def sort_callbacks_(self, reverse: bool = False) -> CallbacksList:
        """
        Sort the callbacks by their priority. The higher the priority, the earlier the callback is called. In general,
        the callbacks will be sorted in the following order:
            1. TrainingHistory callbacks;
            2. Others callbacks;
            3. CheckpointManager callbacks.

        :param reverse: Whether to reverse the order of the callbacks. Default is False.
        :type reverse: bool
        :return: The sorted callbacks.
        :rtype: CallbacksList
        """
        self.callbacks.sort_callbacks_(reverse=reverse)
        return self.callbacks

    def load_state(self):
        """
        Load the state of the trainer from the checkpoint.
        """
        if self.checkpoint_managers:
            main_checkpoint_manager: CheckpointManager = self.checkpoint_managers[0]
            checkpoint = main_checkpoint_manager.curr_checkpoint
            if checkpoint:
                self.callbacks.load_checkpoint_state(self, checkpoint)

    def train(
            self,
            train_dataloader: DataLoader,
            val_dataloader: Optional[DataLoader] = None,
            n_iterations: Optional[int] = None,
            *,
            n_epochs: int = 1,
            load_checkpoint_mode: LoadCheckpointMode = None,
            force_overwrite: bool = False,
            p_bar_position: Optional[int] = None,
            p_bar_leave: Optional[bool] = None,
            **kwargs
    ) -> TrainingHistory:
        """
        Train the model.

        :param train_dataloader: The dataloader for the training set. It contains the training data.
        :type train_dataloader: DataLoader
        :param val_dataloader: The dataloader for the validation set. It contains the validation data.
        :type val_dataloader: Optional[DataLoader]
        :param n_iterations: The number of iterations to train the model. An iteration is a pass over the training set
            and the validation set. If None, the model will be trained until the training is stopped by the user.
        :type n_iterations: Optional[int]
        :param n_epochs: The number of epochs to train the model. An epoch is a pass over the training set. The
            nomenclature here is different from what is usually used elsewhere. Here, an epoch is a pass over the
            training set, while an iteration is a pass over the training set and the validation set. In other words,
            if `n_iterations=1` and `n_epochs=10`, the trainer will pass 10 times over the training set  and 1 time
            over the validation set (this will constitute 1 iteration). If `n_iterations=10` and `n_epochs=1`, the
            trainer will pass 10 times over the training set and 10 times over the validation set (this will constitute
            10 iterations). The nuance between those terms is really important when is comes to reinforcement learning.
            Default is 1.
        :type n_epochs: int
        :param load_checkpoint_mode: The mode to use when loading the checkpoint.
        :type load_checkpoint_mode: LoadCheckpointMode
        :param force_overwrite: Whether to force overwriting the checkpoint. Be careful when using this option, as it
            will destroy the previous checkpoint folder. Default is False.
        :type force_overwrite: bool
        :param p_bar_position: The position of the progress bar. See tqdm documentation for more information.
        :type p_bar_position: Optional[int]
        :param p_bar_leave: Whether to leave the progress bar. See tqdm documentation for more information.
        :type p_bar_leave: Optional[bool]
        :param kwargs: Additional keyword arguments.

        :return: The training history.
        """
        self._load_checkpoint_mode = load_checkpoint_mode
        self._force_overwrite = force_overwrite
        self.kwargs.update(kwargs)
        self.update_state_(
            n_iterations=n_iterations,
            n_epochs=n_epochs,
            objects={
                **self.current_training_state.objects,
                **{"train_dataloader": train_dataloader, "val_dataloader": val_dataloader}
            }
        )
        self.sort_callbacks_()
        self.callbacks.start(self)
        self.load_state()
        if self.current_training_state.iteration is None:
            self.update_state_(iteration=0)
        if len(self.training_history) > 0:
            self.update_itr_metrics_state_(**self.training_history.get_item_at(-1))
        else:
            self.update_state_(itr_metrics={})
        p_bar = tqdm(
            initial=self.current_training_state.iteration,
            total=self.current_training_state.n_iterations,
            desc=kwargs.get("desc", "Training"),
            disable=not self.verbose,
            position=p_bar_position,
            unit="itr",
            leave=p_bar_leave
        )
        self.update_objects_state_(p_bar=p_bar)
        for i in self._iterations_generator(p_bar):
            self.update_state_(iteration=i)
            self.callbacks.on_iteration_begin(self)
            train_dataloader = self.current_training_state.objects["train_dataloader"]
            val_dataloader = self.current_training_state.objects["val_dataloader"]
            itr_loss = self._exec_iteration(train_dataloader, val_dataloader)
            if self.kwargs["exec_metrics_on_train"]:
                itr_train_metrics = self._exec_metrics(train_dataloader, prefix="train")
            else:
                itr_train_metrics = {}
            if val_dataloader is not None:
                itr_val_metrics = self._exec_metrics(val_dataloader, prefix="val")
            else:
                itr_val_metrics = {}
            self.update_itr_metrics_state_(**dict(**itr_loss, **itr_train_metrics, **itr_val_metrics))
            postfix = {f"{k}": f"{v:.5e}" for k, v in self.state.itr_metrics.items()}
            postfix.update(self.callbacks.on_pbar_update(self))
            self.callbacks.on_iteration_end(self)
            p_bar.set_postfix(postfix)
            if self.current_training_state.stop_training_flag:
                p_bar.set_postfix(OrderedDict(**{"stop_flag": "True"}, **postfix))
                break
        self.callbacks.close(self)
        p_bar.close()
        return self.training_history

    def _iterations_generator(self, p_bar: tqdm) -> Generator:
        """
        Generator that yields the current iteration and updates the state and the p_bar.

        :return: The current iteration.
        """
        while self.current_training_state.iteration < self._get_numeric_n_iterations():
            yield self.current_training_state.iteration
            self.update_state_(iteration=self.current_training_state.iteration + 1)
            p_bar.total = self.current_training_state.n_iterations
            p_bar.update()

    def _get_numeric_n_iterations(self) -> int:
        """
        Returns the number of iterations.

        :return: The number of iterations.
        """
        n_iterations = self.current_training_state.n_iterations
        if n_iterations is None:
            n_iterations = np.inf
        return n_iterations

    def _exec_iteration(
            self,
            train_dataloader: DataLoader,
            val_dataloader: Optional[DataLoader] = None
    ) -> Dict[str, float]:
        with torch.no_grad():
            torch.cuda.empty_cache()
        losses = {}

        self.model.train()
        self.callbacks.on_train_begin(self)
        self.update_state_(batch_is_train=True)
        train_losses = []
        for epoch_idx in range(self.current_training_state.n_epochs):
            self.update_state_(epoch=epoch_idx)
            train_losses.append(self._exec_epoch(train_dataloader))
        train_loss = np.mean(train_losses)
        self.update_state_(train_loss=train_loss)
        self.callbacks.on_train_end(self)
        losses["train_loss"] = train_loss

        if val_dataloader is not None:
            with torch.no_grad():
                self.model.eval()
                self.callbacks.on_validation_begin(self)
                self.update_state_(batch_is_train=False)
                val_loss = self._exec_epoch(val_dataloader)
                self.update_state_(val_loss=val_loss)
                self.callbacks.on_validation_end(self)
                losses["val_loss"] = val_loss

        with torch.no_grad():
            torch.cuda.empty_cache()
        return losses

    def _exec_metrics(self, dataloader: torch.utils.data.DataLoader, prefix: str) -> Dict:
        metrics_dict = {}
        for metric in self.metrics:
            m_out = metric(dataloader)
            if isinstance(m_out, dict):
                metrics_dict.update({f"{prefix}_{k}": v for k, v in m_out.items()})
            else:
                metric_name = str(metric)
                if hasattr(metric, "name"):
                    metric_name = metric.name
                elif hasattr(metric, "__name__"):
                    metric_name = metric.__name__
                metrics_dict[f"{prefix}_{metric_name}"] = m_out
        return metrics_dict

    def _exec_epoch(
            self,
            dataloader: DataLoader,
    ) -> float:
        self.callbacks.on_epoch_begin(self)
        batch_losses = []
        for i, entries in enumerate(dataloader):
            x_batch, hh_batch, y_batch = unpack_x_hh_y(entries)
            self.update_state_(batch=i)
            batch_losses.append(to_numpy(self._exec_batch(x_batch, hh_batch, y_batch)))
        mean_loss = np.mean(batch_losses)
        self.callbacks.on_epoch_end(self)
        return mean_loss

    def _exec_batch(
            self,
            x_batch,
            hh_batch,
            y_batch,
    ):
        x_batch = self.x_transform(self._batch_to_dense(self._batch_to_device(x_batch)))
        if hh_batch is not None:
            hh_batch = self.x_transform(self._batch_to_dense(self._batch_to_device(hh_batch)))
        y_batch = self.y_transform(self._batch_to_dense(self._batch_to_device(y_batch)))
        self.update_state_(x_batch=x_batch, hh_batch=hh_batch, y_batch=y_batch)
        self.callbacks.on_batch_begin(self)
        pred_batch = self.get_pred_batch(x_batch, hh_batch)
        self.update_state_(pred_batch=pred_batch)
        if self.model.training:
            self.callbacks.on_optimization_begin(self, x=x_batch, hh_batch=hh_batch, y=y_batch, pred=pred_batch)
            self.callbacks.on_optimization_end(self)
        else:
            self.callbacks.on_validation_batch_begin(self, x=x_batch, hh_batch=hh_batch, y=y_batch, pred=pred_batch)
            self.callbacks.on_validation_batch_end(self)
        self.callbacks.on_batch_end(self)
        batch_loss = self.current_training_state.batch_loss
        if batch_loss is None:
            batch_loss = 0.0
        else:
            batch_loss = batch_loss.item()
        return batch_loss

    def get_pred_batch(
            self,
            x_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
            hh_batch: Optional[Any] = None,
            *args,
            **kwargs
    ):
        """
        Get the prediction of the model on the given batch.

        :param x_batch: The input batch.
        :type x_batch: Union[torch.Tensor, Dict[str, torch.Tensor]]
        :param hh_batch: The hidden state batch.
        :type hh_batch: Optional[Any]
        :param args: Additional arguments to pass to the model.
        :param kwargs: Additional keyword arguments to pass to the model.
        """
        if hh_batch is not None:
            args = (hh_batch, *args)
        if self.model.training:
            out = getattr(self.model, self.predict_method)(x_batch, *args, **kwargs)
        else:
            with torch.no_grad():
                out = getattr(self.model, self.predict_method)(x_batch, *args, **kwargs)
        return out

    def apply_criterion_on_batch(
            self,
            x_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
            y_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
            pred_batch: Optional[Union[torch.Tensor, Dict[str, torch.Tensor]]]
    ) -> torch.Tensor:
        if isinstance(pred_batch, (tuple, list)):
            pred = pred_batch[0]
        elif isinstance(pred_batch, torch.Tensor):
            pred = pred_batch
        else:
            raise ValueError(f"Unsupported output type: {type(pred_batch)}")

        if isinstance(self.criterion, dict):
            if isinstance(y_batch, torch.Tensor):
                y_batch = {k: y_batch for k in self.criterion}
            if isinstance(pred, torch.Tensor):
                pred = {k: pred for k in self.criterion}
            assert isinstance(pred, dict) and isinstance(y_batch, dict) and isinstance(pred, dict), \
                "If criterion is a dict, pred, y_batch and pred must be a dict too."
            batch_loss = sum([
                self.criterion[k](pred[k], y_batch[k].to(self.device))
                for k in self.criterion
            ])
        else:
            if isinstance(pred, dict) and len(pred) == 1:
                pred = pred[list(pred.keys())[0]]
            batch_loss = self.criterion(pred, y_batch.to(self.device))
        return batch_loss

    def _batch_to_dense(self, batch):
        if isinstance(batch, dict):
            return {k: self._batch_to_dense(v) for k, v in batch.items()}
        if isinstance(batch, torch.Tensor) and batch.is_sparse:
            return batch.to_dense()
        return batch

    def _batch_to_device(self, batch):
        if isinstance(batch, dict):
            return {k: self._batch_to_device(v) for k, v in batch.items()}
        if isinstance(batch, torch.Tensor):
            return batch.to(self.device, non_blocking=True)
        return batch

    def __repr__(self):
        repr_ = f"{self.__class__.__name__}("
        repr_ += f"\n\tmodel={self.model}, "
        repr_ += f"\n\tpredict_method={self.predict_method}, "
        repr_ += f"\n\tcallbacks={self.callbacks}"
        repr_ += f")@{self.device}"
        return repr_






import json
import logging
from copy import deepcopy
from typing import Sequence, Union, Optional, Dict, Any, List

import numpy as np
import torch
import gym

from ..callbacks.checkpoints_manager import CheckpointManager, LoadCheckpointMode
from ..transforms.base import to_numpy, to_tensor
from ..modules.base import BaseModel
from ..modules.sequential import Sequential
from ..utils import maybe_apply_softmax, unpack_out_hh

from ..modules.layers import Linear
from .utils import (
    obs_sequence_to_batch,
    space_to_spec,
    obs_batch_to_sequence,
    space_to_continuous_shape,
    get_single_observation_space,
    get_single_action_space,
    sample_action_space,
    continuous_actions_distribution,
)


class Agent(torch.nn.Module):
    @staticmethod
    def copy_from_agent(agent: "Agent", requires_grad: Optional[bool] = None) -> "Agent":
        """
        Copy the agent.

        :param agent: The agent to copy.
        :type agent: Agent
        :param requires_grad: Whether to require gradients.
        :type requires_grad: Optional[bool]
        :return: The copied agent.
        :rtype: Agent
        """
        return Agent(
            env=agent.env,
            observation_space=agent.observation_space,
            action_space=agent.action_space,
            behavior_name=agent.behavior_name,
            policy=agent.copy_policy(requires_grad=requires_grad),
            critic=agent.copy_critic(requires_grad=requires_grad),
            **agent.kwargs
        )

    def __init__(
            self,
            *,
            env: Optional[gym.Env] = None,
            observation_space: Optional[gym.spaces.Space] = None,
            action_space: Optional[gym.spaces.Space] = None,
            behavior_name: Optional[str] = None,
            policy: Optional[BaseModel] = None,
            policy_predict_method: str = "__call__",
            policy_kwargs: Optional[Dict[str, Any]] = None,
            critic: Optional[BaseModel] = None,
            critic_predict_method: str = "__call__",
            critic_kwargs: Optional[Dict[str, Any]] = None,
            **kwargs
    ):
        """
        Constructor for BaseAgent class.

        :param env: The environment.
        :type env: Optional[gym.Env]
        :param observation_space: The observation space. Must be a single space not batched. Must be provided if
            `env` is not provided. If `env` is provided, then this will be ignored.
        :type observation_space: Optional[gym.spaces.Space]
        :param action_space: The action space. Must be a single space not batched. Must be provided if
            `env` is not provided. If `env` is provided, then this will be ignored.
        :type action_space: Optional[gym.spaces.Space]
        :param behavior_name: The name of the behavior.
        :type behavior_name: Optional[str]
        :param policy: The model to use.
        :type policy: BaseModel
        :param policy_kwargs: The keyword arguments to pass to the policy if it is created by default.
            The keywords are:
                - `default_hidden_units` (List[int]): The default number of hidden units. Defaults to [256].
                - `default_activation` (str): The default activation function. Defaults to "ReLu".
                - `default_output_activation` (str): The default output activation function. Defaults to "Identity".
                - `default_dropout` (float): The default dropout rate. Defaults to 0.1.
                - all other keywords are passed to the `Sequential` constructor.
        :type policy_kwargs: Optional[Dict[str, Any]]
        :param critic: The value model to use.
        :type critic: BaseModel
        :param critic_kwargs: The keyword arguments to pass to the critic if it is created by default.
            The keywords are:
                - `default_hidden_units` (List[int]): The default number of hidden units. Defaults to [256].
                - `default_activation` (str): The default activation function. Defaults to "ReLu".
                - `default_output_activation` (str): The default output activation function. Defaults to "Identity".
                - `default_n_values` (int): The default number of values to output. Defaults to 1.
                - `default_dropout` (float): The default dropout rate. Defaults to 0.1.
                - all other keywords are passed to the `Sequential` constructor.
        :type critic_kwargs: Optional[Dict[str, Any]]
        :param kwargs: Other keyword arguments.
        """
        super().__init__()
        self.kwargs = kwargs
        self.policy_kwargs = policy_kwargs if policy_kwargs is not None else {}
        self.set_default_policy_kwargs()
        self.critic_kwargs = critic_kwargs if critic_kwargs is not None else {}
        self.set_default_critic_kwargs()
        self.env = env
        if env:
            self.observation_space = get_single_observation_space(env)
            self.action_space = get_single_action_space(env)
        else:
            self.observation_space = observation_space
            self.action_space = action_space
        if behavior_name:
            self.behavior_name = behavior_name
        elif env.spec:
            self.behavior_name = env.spec.id
        else:
            self.behavior_name = "default"
        self.policy = policy
        if self.policy is None:
            self.policy = self._create_default_policy()
        self.policy_predict_method_name = policy_predict_method
        assert hasattr(self.policy, self.policy_predict_method_name), \
            f"Policy does not have method '{self.policy_predict_method_name}'"
        self.policy_predict_method = getattr(self.policy, self.policy_predict_method_name)
        assert callable(self.policy_predict_method), \
            f"Policy method '{self.policy_predict_method_name}' is not callable"
        self.critic = critic
        if self.critic is None:
            self.critic = self._create_default_critic()
        self.critic_predict_method_name = critic_predict_method
        assert hasattr(self.critic, self.critic_predict_method_name), \
            f"Critic does not have method '{self.critic_predict_method_name}'"
        self.critic_predict_method = getattr(self.critic, self.critic_predict_method_name)
        assert callable(self.policy_predict_method), \
            f"Critic method '{self.critic_predict_method_name}' is not callable"
        self.checkpoint_folder = kwargs.get("checkpoint_folder", ".")

        self.checkpoints_meta_path = kwargs.get("checkpoints_meta_path", self.get_default_checkpoints_meta_path())
        continuous_action_mid_ranges = {
            k: to_tensor(self.action_spec[k].high - self.action_spec[k].low) / 2
            for k in self.continuous_actions
        }
        if len(continuous_action_mid_ranges) > 0:
            continuous_action_mid_ranges_mean = torch.mean(torch.stack(list(continuous_action_mid_ranges.values())))
        else:
            continuous_action_mid_ranges_mean = 0.0
        self.continuous_action_variances = torch.nn.ParameterDict(
            {  # TODO: Make this a parameter and the keys must be the same or mapped with the output layers names
                k: continuous_action_mid_ranges[k] * torch.ones(self.action_spec[k].shape)
                for k in self.continuous_actions
            }
        )
        self.continuous_action_variances_decay = to_tensor(kwargs.get("continuous_action_variances_decay", 1 - 1e-4))
        self.continuous_action_variances_min = to_tensor(
            kwargs.get(
                "continuous_action_variances_min", 0.1 * continuous_action_mid_ranges_mean
            )
        )

    @property
    def observation_spec(self) -> Dict[str, Any]:
        return space_to_spec(self.observation_space)

    @property
    def action_spec(self) -> Dict[str, Any]:
        return space_to_spec(self.action_space)

    @property
    def discrete_actions(self) -> List[str]:
        return [k for k, v in self.action_spec.items() if isinstance(v, gym.spaces.Discrete)]

    @property
    def continuous_actions(self) -> List[str]:
        return [k for k, v in self.action_spec.items() if not isinstance(v, gym.spaces.Discrete)]

    @property
    def device(self) -> torch.device:
        """
        The device of the agent.

        :return: The device of the agent.
        :rtype: torch.device
        """
        return next(self.parameters()).device

    @device.setter
    def device(self, device: torch.device):
        """
        Set the device of the agent.

        :param device: The device to set.
        :type device: torch.device
        """
        self.policy.to(device)
        if self.critic is not None:
            self.critic.to(device)

    def get_continuous_action_covariances(self):
        return {
            k: torch.diag(v) for k, v in self.continuous_action_variances.items()
        }

    def decay_continuous_action_variances(self):
        for k in self.continuous_action_variances:
            self.continuous_action_variances[k] = torch.clamp(
                self.continuous_action_variances[k] * self.continuous_action_variances_decay,
                min=self.continuous_action_variances_min,
                max=torch.inf,
                )

    def set_continuous_action_variances_with_itr(self, itr: int):
        for k in self.continuous_action_variances:
            self.continuous_action_variances[k] = torch.clamp(
                self.continuous_action_variances[k] * (self.continuous_action_variances_decay**itr),
                min=self.continuous_action_variances_min,
                max=torch.inf,
                )

    def get_default_checkpoints_meta_path(self) -> str:
        """
        The path to the checkpoints meta file.

        :return: The path to the checkpoints meta file.
        :rtype: str
        """
        full_filename = (
            f"{self.behavior_name}{CheckpointManager.SUFFIX_SEP}{CheckpointManager.CHECKPOINTS_META_SUFFIX}"
        )
        return f"{self.checkpoint_folder}/{full_filename}.json"

    def set_default_policy_kwargs(self):
        self.policy_kwargs.setdefault("default_hidden_units", [256])
        if isinstance(self.policy_kwargs["default_hidden_units"], int):
            self.policy_kwargs["default_hidden_units"] = [self.policy_kwargs["default_hidden_units"]]
        assert len(self.policy_kwargs["default_hidden_units"]) > 0, "Must have at least one hidden unit."
        self.policy_kwargs.setdefault("default_activation", "ReLu")
        self.policy_kwargs.setdefault("default_output_activation", "Identity")
        self.policy_kwargs.setdefault("default_dropout", 0.1)

    def set_default_critic_kwargs(self):
        self.critic_kwargs.setdefault("default_hidden_units", [256])
        if isinstance(self.critic_kwargs["default_hidden_units"], int):
            self.critic_kwargs["default_hidden_units"] = [self.critic_kwargs["default_hidden_units"]]
        assert len(self.critic_kwargs["default_hidden_units"]) > 0, "Must have at least one hidden unit."
        self.critic_kwargs.setdefault("default_activation", "ReLu")
        self.critic_kwargs.setdefault("default_output_activation", "Identity")
        self.critic_kwargs.setdefault("default_n_values", 1)
        self.critic_kwargs.setdefault("default_dropout", 0.1)

    def _create_default_policy(self) -> BaseModel:
        """
        Create the default policy.

        :return: The default policy.
        :rtype: BaseModel
        """
        hidden_block = [torch.nn.Dropout(p=self.policy_kwargs["default_dropout"])]
        for i in range(len(self.policy_kwargs["default_hidden_units"]) - 1):
            hidden_block.extend([
                Linear(
                    input_size=self.policy_kwargs["default_hidden_units"][i],
                    output_size=self.policy_kwargs["default_hidden_units"][i + 1],
                    activation=self.policy_kwargs["default_activation"]
                ),
                # torch.nn.Linear(
                # 	in_features=self.policy_kwargs["default_hidden_units"][i],
                # 	out_features=self.policy_kwargs["default_hidden_units"][i + 1]
                # ),
                # torch.nn.PReLU(),  # TODO: for Debugging
                torch.nn.Dropout(p=self.policy_kwargs["default_dropout"]),
            ])
        default_policy = Sequential(layers=[
            {
                f"in_{k}": Linear(
                    input_size=int(space_to_continuous_shape(v, flatten_spaces=True)[0]),
                    output_size=self.policy_kwargs["default_hidden_units"][0],
                    activation=self.policy_kwargs["default_activation"]
                )
                # k: torch.nn.Linear(
                # 	in_features=int(space_to_continuous_shape(v, flatten_spaces=True)[0]),
                # 	out_features=self.policy_kwargs["default_hidden_units"][0]
                # )
                for k, v in self.observation_spec.items()
            },
            *hidden_block,
            {
                f"out_{k}": Linear(
                    input_size=self.policy_kwargs["default_hidden_units"][-1],
                    output_size=int(space_to_continuous_shape(v, flatten_spaces=True)[0]),
                    activation=self.policy_kwargs["default_output_activation"]
                )
                for k, v in self.action_spec.items()
            }
        ],
            **self.policy_kwargs
        ).build()
        return default_policy

    def _create_default_critic(self) -> BaseModel:
        """
        Create the default critic.

        :return: The default critic.
        :rtype: BaseModel
        """
        hidden_block = [torch.nn.Dropout(p=self.critic_kwargs["default_dropout"])]
        for i in range(len(self.policy_kwargs["default_hidden_units"]) - 1):
            hidden_block.extend([
                Linear(
                    input_size=self.critic_kwargs["default_hidden_units"][i],
                    output_size=self.critic_kwargs["default_hidden_units"][i + 1],
                    activation=self.critic_kwargs["default_activation"]
                ),
                torch.nn.Dropout(p=self.critic_kwargs["default_dropout"]),
            ])
        default_policy = Sequential(layers=[
            {
                f"in_{k}": Linear(
                    input_size=int(space_to_continuous_shape(v, flatten_spaces=True)[0]),
                    output_size=self.critic_kwargs["default_hidden_units"][0],
                    activation=self.critic_kwargs["default_activation"]
                )
                for k, v in self.observation_spec.items()
            },
            *hidden_block,
            Linear(
                input_size=self.critic_kwargs["default_hidden_units"][-1],
                output_size=self.critic_kwargs["default_n_values"],
                activation=self.critic_kwargs["default_output_activation"]
            )
        ],
            **self.critic_kwargs
        ).build()
        return default_policy

    def forward(self, *args, **kwargs):
        """
        Call the agent.

        :return: The output of the agent.
        """
        return self.policy_predict_method(*args, **kwargs)

    def get_actions(
            self,
            obs: Union[np.ndarray, torch.Tensor, Dict[str, Union[np.ndarray, torch.Tensor]]],
            **kwargs
    ) -> Any:
        """
        Get the actions for the given observations.

        :param obs: The observations. The observations must be batched.
        :type obs: Union[np.ndarray, torch.Tensor, Dict[str, Union[np.ndarray, torch.Tensor]]]
        :param kwargs: Keywords arguments.

        :keyword str re_format: The format to reformat the discrete actions to. Default is "index" which
            will return the index of the action. For other options see :mth:`format_batch_discrete_actions`.
        :keyword bool as_numpy: Whether to return the actions as numpy arrays. Default is True.

        :return: The actions.
        """
        self.env = kwargs.get("env", self.env)
        re_as_dict = kwargs.get("re_as_dict", isinstance(obs, dict) or isinstance(obs[0], dict))
        re_formats = kwargs.get("re_format", "index").split(",")
        as_numpy = kwargs.get("as_numpy", True)

        obs_as_tensor = to_tensor(obs)
        out_actions, _ = unpack_out_hh(self.policy_predict_method(obs_as_tensor, **kwargs))
        re_actions_list = [
            self.format_batch_discrete_actions(out_actions, re_format=re_format)
            for re_format in re_formats
        ]
        if as_numpy:
            re_actions_list = [to_numpy(re_actions) for re_actions in re_actions_list]
        for i, re_actions in enumerate(re_actions_list):
            if not re_as_dict and isinstance(re_actions, dict):
                if not len(re_actions) == 1:
                    raise ValueError("Cannot unpack actions from dict because it has not a length of 1.")
                re_actions_list[i] = re_actions[list(re_actions.keys())[0]]
            elif re_as_dict and not isinstance(re_actions, dict):
                keys = self.discrete_actions + self.continuous_actions
                re_actions_list[i] = {k: re_actions for k in keys}
        if len(re_actions_list) == 1:
            return re_actions_list[0]
        return re_actions_list

    def format_batch_discrete_actions(
            self,
            actions: Union[torch.Tensor, Dict[str, torch.Tensor]],
            re_format: str = "logits",
            **kwargs
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        """
        Format the batch of actions. If actions is a dict, then it is assumed that the keys are the action names and the
        values are the actions. In this case, all the values where their keys are in `self.discrete_actions` will be
        formatted. If actions is a tensor, then the actions will be formatted if `self.discrete_actions` is not empty.

        TODO: fragment this method into smaller methods.

        :param actions: The actions.
        :param re_format: The format to reformat the actions to. Can be "logits", "probs", "index", or "one_hot".
        :param kwargs: Keywords arguments.
        :return: The formatted actions.
        """
        discrete_actions = kwargs.get("discrete_actions", self.discrete_actions)
        continuous_actions = kwargs.get("continuous_actions", self.continuous_actions)
        actions = to_tensor(actions)
        if re_format.lower() in ["logits", "raw"]:
            return actions
        elif re_format.lower() == "probs":
            if isinstance(actions, torch.Tensor):
                return torch.softmax(actions, dim=-1) if len(discrete_actions) >= 1 else actions
            elif isinstance(actions, dict):
                return {k: (
                    torch.softmax(v, dim=-1) if (k in discrete_actions or len(continuous_actions) == 0) else v
                ) for k, v in actions.items()}
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() == "log_probs":
            if isinstance(actions, torch.Tensor):
                return torch.log_softmax(actions, dim=-1) if len(discrete_actions) >= 1 else actions
            elif isinstance(actions, dict):
                return {k: (
                    torch.log_softmax(v, dim=-1)
                    if (k in discrete_actions or len(continuous_actions) == 0) else v
                ) for k, v in actions.items()}
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() in ["index", "indices", "argmax", "imax", "amax"]:
            if isinstance(actions, torch.Tensor):
                return torch.argmax(actions, dim=-1).long() if len(discrete_actions) >= 1 else actions
            elif isinstance(actions, dict):
                return {k: (
                    torch.argmax(v, dim=-1).long()
                    if (k in discrete_actions or len(continuous_actions) == 0) else v
                ) for k, v in actions.items()}
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() == "one_hot":
            if isinstance(actions, torch.Tensor):
                return (
                    torch.nn.functional.one_hot(torch.argmax(actions, dim=-1), num_classes=actions.shape[-1])
                    if len(discrete_actions) >= 1 else actions
                )
            elif isinstance(actions, dict):
                return {
                    k: (
                        torch.nn.functional.one_hot(torch.argmax(v, dim=-1), num_classes=v.shape[-1])
                        if (k in discrete_actions or len(continuous_actions) == 0) else v
                    )
                    for k, v in actions.items()
                }
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() == "max":
            if isinstance(actions, torch.Tensor):
                return torch.max(actions, dim=-1).values if len(discrete_actions) >= 1 else actions
            elif isinstance(actions, dict):
                return {k: (
                    torch.max(v, dim=-1).values
                    if (k in discrete_actions or len(continuous_actions) == 0) else v
                ) for k, v in actions.items()}
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() == "smax":
            if isinstance(actions, torch.Tensor):
                return torch.softmax(actions, dim=-1).max(dim=-1).values if len(discrete_actions) >= 1 else actions
            elif isinstance(actions, dict):
                return {
                    k: (
                        torch.softmax(v, dim=-1).max(dim=-1).values
                        if (k in discrete_actions or len(continuous_actions) == 0) else v
                    )
                    for k, v in actions.items()
                }
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() == "log_smax":
            if isinstance(actions, torch.Tensor):
                return torch.log_softmax(actions, dim=-1).max(dim=-1).values if len(discrete_actions) >= 1 else actions
            elif isinstance(actions, dict):
                return {
                    k: (
                        torch.log_softmax(v, dim=-1).max(dim=-1).values
                        if (k in discrete_actions or len(continuous_actions) == 0) else v
                    )
                    for k, v in actions.items()
                }
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        elif re_format.lower() == "sample":
            if isinstance(actions, torch.Tensor):
                if len(discrete_actions) >= 1:
                    probs = maybe_apply_softmax(actions, dim=-1)
                    return torch.distributions.Categorical(probs=probs).sample()
                else:
                    covariance = self.get_continuous_action_covariances()[self.continuous_actions[0]]
                    return continuous_actions_distribution(actions, covariance=covariance).sample()
            elif isinstance(actions, dict):
                return {
                    k: (
                        torch.distributions.Categorical(probs=maybe_apply_softmax(v, dim=-1)).sample()
                        if (k in discrete_actions or len(continuous_actions) == 0)
                        else continuous_actions_distribution(
                            # TODO: must get the right covariance for each continuous action
                            v, covariance=self.get_continuous_action_covariances()[self.continuous_actions[0]]
                        ).sample()
                    )
                    for k, v in actions.items()
                }
            else:
                raise ValueError(f"Cannot format actions of type {type(actions)}.")
        else:
            raise ValueError(f"Unknown re-formatting option {re_format}.")

    def get_random_actions(self, n_samples: int = 1, **kwargs) -> Any:
        as_batch = kwargs.get("as_batch", False)
        as_sequence = kwargs.get("as_sequence", False)
        re_formats = kwargs.get("re_format", "index").split(",")
        as_numpy = kwargs.get("as_numpy", True)
        assert not (as_batch and as_sequence), "Cannot return actions as both batch and sequence."
        as_single = not (as_batch or as_sequence)
        self.env = kwargs.get("env", self.env)
        if self.env:
            action_space = self.env.action_space
        else:
            action_space = self.action_space
        if as_single and n_samples == 1:
            out_actions = sample_action_space(action_space, re_format="one_hot")
        else:
            out_actions = [sample_action_space(action_space, re_format="one_hot") for _ in range(n_samples)]
            if as_batch:
                out_actions = obs_sequence_to_batch(out_actions)
        re_actions_list = [
            self.format_batch_discrete_actions(out_actions, re_format=re_format)
            for re_format in re_formats
        ]
        if as_numpy:
            re_actions_list = [to_numpy(re_actions) for re_actions in re_actions_list]
        if len(re_actions_list) == 1:
            return re_actions_list[0]
        return re_actions_list

    def get_values(self, obs: torch.Tensor, **kwargs) -> Any:
        """
        Get the values for the given observations.

        :param obs: The batched observations.
        :param kwargs: Keywords arguments.
        :return: The values.
        """
        self.env = kwargs.get("env", self.env)
        re_as_dict = kwargs.get("re_as_dict", isinstance(obs, dict) or isinstance(obs[0], dict))
        as_numpy = kwargs.get("as_numpy", True)
        obs_as_tensor = to_tensor(obs)
        values, _ = unpack_out_hh(self.critic_predict_method(obs_as_tensor))
        if as_numpy:
            values = to_numpy(values)
        if not re_as_dict and isinstance(values, dict):
            values = values[list(values.keys())[0]]
        return values

    def __str__(self):
        n_tab = 2
        policy_repr = str(self.policy)
        tab_policy_repr = "\t" + policy_repr.replace("\n", "\n"+("\t"*n_tab))
        critic_repr = str(self.critic)
        tab_critic_repr = "\t" + critic_repr.replace("\n", "\n"+("\t"*n_tab))
        agent_repr = f"Agent<{self.behavior_name}>:\n\t[Policy](\n{tab_policy_repr}\t\n)\n"
        if self.critic:
            agent_repr += f"\t[Critic](\n{tab_critic_repr}\t\n)\n"
        return agent_repr

    def soft_update(self, policy, tau):
        self.policy.soft_update(policy, tau)

    def hard_update(self, policy):
        self.policy.hard_update(policy)

    def copy(self, requires_grad: Optional[bool] = None) -> "Agent":
        """
        Copy the agent.

        :param requires_grad: Whether to require gradients.
        :type requires_grad: Optional[bool]
        :return: The copied agent.
        :rtype: Agent
        """
        return self.copy_from_agent(self, requires_grad=requires_grad)

    def copy_critic(self, requires_grad: Optional[bool] = None) -> BaseModel:
        """
        Copy the critic to a new instance.

        :return: The copied critic.
        """
        critic_copy = deepcopy(self.critic)
        if requires_grad is not None:
            for param in critic_copy.parameters():
                param.requires_grad = requires_grad
        return critic_copy

    def copy_policy(self, requires_grad: Optional[bool] = None) -> BaseModel:
        """
        Copy the policy to a new instance.

        :return: The copied policy.
        """
        policy_copy = deepcopy(self.policy)
        if requires_grad is not None:
            for param in policy_copy.parameters():
                param.requires_grad = requires_grad
        return policy_copy

    def load_checkpoint(
            self,
            checkpoints_meta_path: Optional[str] = None,
            load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR,
            verbose: bool = True
    ) -> dict:
        """
        Load the checkpoint from the checkpoints_meta_path. If the checkpoints_meta_path is None, the default
        checkpoints_meta_path is used.

        :param checkpoints_meta_path: The path to the checkpoints meta file.
        :type checkpoints_meta_path: Optional[str]
        :param load_checkpoint_mode: The mode to use when loading the checkpoint.
        :type load_checkpoint_mode: LoadCheckpointMode
        :param verbose: Whether to print the loaded checkpoint information.
        :type verbose: bool

        :return: The loaded checkpoint information.
        :rtype: dict
        """
        if checkpoints_meta_path is None:
            checkpoints_meta_path = self.checkpoints_meta_path
        with open(checkpoints_meta_path, "r+") as jsonFile:
            info: dict = json.load(jsonFile)
        save_name = CheckpointManager.get_save_name_from_checkpoints(info, load_checkpoint_mode)
        checkpoint_path = f"{self.checkpoint_folder}/{save_name}"
        if verbose:
            logging.info(f"Loading checkpoint from {checkpoint_path}")
        checkpoint = torch.load(checkpoint_path, map_location=self.device)
        self.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_STATE_DICT_KEY], strict=True)
        return checkpoint

    def to(self, *args, **kwargs):
        self.policy.to(*args, **kwargs)
        if self.critic is not None:
            self.critic.to(*args, **kwargs)
        return self











import time
from typing import Optional

import numpy as np

from .base_callback import BaseCallback


class EarlyStopping(BaseCallback):
    def __init__(
            self,
            patience: int = 5,
            tol: float = 1e-2,
    ):
        self.patience = patience
        self.tol = tol
        self.best_loss = None
        self.counter = 0

    def _check_early_stopping(self, patience: int, tol: float = 1e-2) -> bool:
        """
        :param patience:
        :return:
        """
        losses = self.loss_history['val'][-patience:]
        return np.all(np.abs(np.diff(losses)) < tol)


class EarlyStoppingThreshold(BaseCallback):
    """
    Monitor the training process and set the stop_training_flag to True when the threshold is met.
    """
    def __init__(
            self,
            *,
            metric: str,
            threshold: float,
            minimize_metric: bool,
            **kwargs
    ):
        """
        Constructor for EarlyStoppingThreshold class.

        :param metric: Name of the metric to monitor.
        :type metric: str
        :param threshold: Threshold value for the metric.
        :type threshold: float
        :param minimize_metric: Whether to minimize or maximize the metric.
        :type minimize_metric: bool
        :param kwargs: The keyword arguments to pass to the BaseCallback.
        """
        super().__init__(**kwargs)
        self.threshold = threshold
        self.metric = metric
        self.minimize_metric = minimize_metric

    def on_iteration_end(self, trainer, **kwargs):
        if self.minimize_metric:
            threshold_met = trainer.current_training_state.itr_metrics[self.metric] < self.threshold
        else:
            threshold_met = trainer.current_training_state.itr_metrics[self.metric] > self.threshold
        if threshold_met:
            trainer.update_state_(stop_training_flag=True)


class EarlyStoppingOnTimeLimit(BaseCallback):
    """
    Monitor the training process and set the stop_training_flag to True when the threshold is met.
    """
    CURRENT_SECONDS_COUNT_KEY = "current_seconds_count"
    DELTA_SECONDS_KEY = "delta_seconds"

    def __init__(
            self,
            *,
            delta_seconds: float = 10.0 * 60.0,
            resume_on_load: bool = True,
            **kwargs
    ):
        """
        Constructor for EarlyStoppingThreshold class.

        :param delta_seconds: The number of seconds to wait before stopping the training.
        :type delta_seconds: float
        :param resume_on_load: Whether to resume the time when loading a checkpoint. If False, the time will be reset
            to 0.
        :type resume_on_load: bool
        :param kwargs: The keyword arguments to pass to the BaseCallback.
        """
        super().__init__(**kwargs)
        self.delta_seconds = delta_seconds
        self.resume_on_load = resume_on_load
        self.start_time = None
        self.last_time_update = None
        self.current_seconds_count = 0.0

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        self.start_time = None
        self.current_seconds_count = 0.0
        if self.load_state:
            state = checkpoint.get(self.name, {})
            self.delta_seconds = state.get(self.DELTA_SECONDS_KEY, self.delta_seconds)
            if self.resume_on_load:
                self.current_seconds_count = state.get(self.CURRENT_SECONDS_COUNT_KEY, 0.0)

    def get_checkpoint_state(self, trainer, **kwargs) -> object:
        if self.save_state:
            state = {
                self.CURRENT_SECONDS_COUNT_KEY: self.current_seconds_count,
                self.DELTA_SECONDS_KEY: self.delta_seconds
            }
            return state
        return None

    def start(self, trainer, **kwargs):
        self.start_time = time.time()
        self.last_time_update = self.start_time
        self.update_flags(trainer, **kwargs)

    def on_iteration_end(self, trainer, **kwargs):
        self.current_seconds_count += time.time() - self.last_time_update
        self.update_flags(trainer, **kwargs)
        self.last_time_update = time.time()

    def update_flags(self, trainer, **kwargs):
        if self.current_seconds_count > self.delta_seconds:
            trainer.update_state_(stop_training_flag=True)

    def extra_repr(self) -> str:
        _repr = super().extra_repr()
        _repr += f"delta_seconds={self.delta_seconds}"
        return _repr


class EarlyStoppingOnNaN(BaseCallback):
    """
    Monitor the training process and set the stop_training_flag to True when the metric is NaN.
    """
    DEFAULT_PRIORITY = BaseCallback.DEFAULT_LOW_PRIORITY

    def __init__(self, metric: str, **kwargs):
        """
        Constructor for EarlyStoppingOnNaN class.

        :param metric: Name of the metric to monitor.
        :type metric: str
        :param kwargs: The keyword arguments to pass to the BaseCallback.
        """
        super().__init__(**kwargs)
        self.metric = metric

    def on_iteration_end(self, trainer, **kwargs):
        if np.any(np.isnan(trainer.state.itr_metrics.get(self.metric, 0.0))):
            trainer.update_state_(stop_training_flag=True)

    def extra_repr(self) -> str:
        return f"metric: {self.metric}"


class EarlyStoppingOnStagnation(BaseCallback):
    """
    Monitor the training process and set the stop_training_flag to True when the metric stagnates.
    The metric is considered to stagnate when the mean of the absolute difference between the last
    `patience` iterations is less than `tol`.
    """
    DEFAULT_PRIORITY = BaseCallback.DEFAULT_LOW_PRIORITY

    def __init__(
            self,
            metric: str,
            patience: int = 10,
            tol: float = 1e-4,
            start_with_history: bool = True,
            **kwargs
    ):
        """
        Constructor for EarlyStoppingOnStagnation class.

        :param metric: Name of the metric to monitor.
        :type metric: str
        :param patience: Number of iterations to wait before stopping the training.
        :type patience: int
        :param tol: The tolerance for the metric.
        :type tol: float
        :param start_with_history: Whether to start the monitor with the history of the metric.
        :type start_with_history: bool
        :param kwargs: The keyword arguments to pass to the BaseCallback.
        """
        super().__init__(**kwargs)
        self.metric = metric
        self.patience = patience
        if patience < 2:
            raise ValueError("The patience must be at least 2.")
        self.tol = tol
        self.start_with_history = start_with_history
        self._memory = []

    def get_value(self):
        if len(self._memory) < 2:
            return np.inf
        return np.mean(np.abs(np.diff(np.asarray(self._memory))))

    def start(self, trainer, **kwargs):
        self._memory = []
        if not self.start_with_history:
            return
        history = trainer.training_history
        if history is not None:
            self._memory = history[self.metric][-self.patience:]

    def on_iteration_end(self, trainer, **kwargs):
        self._memory.append(trainer.state.itr_metrics.get(self.metric, 0.0))
        if len(self._memory) < self.patience:
            return
        self._memory.pop(0)
        if self.get_value() < self.tol:
            trainer.update_state_(stop_training_flag=True)

    def extra_repr(self) -> str:
        return f"metric: {self.metric}, patience={self.patience}, tol={self.tol}"





import os
import warnings
from collections import defaultdict
from typing import Optional, Union, Sequence, Dict, Tuple, Any, List

import matplotlib.pyplot as plt
import matplotlib.animation as mpl_animation
import numpy as np
import gym
import scipy
import torch

from .buffers import Trajectory
from .curriculum import Curriculum
from ..dimension import SizeTypes
from ..modules.layers import BaseNeuronsLayer
from ..transforms.base import to_tensor, to_numpy
from ..callbacks import TrainingHistory
from ..utils import legend_without_duplicate_labels_


class TrainingHistoriesMap:
    REPORT_KEY = "report"

    def __init__(self, curriculum: Optional[Curriculum] = None):
        self.curriculum = curriculum
        self.histories = defaultdict(TrainingHistory, **{TrainingHistoriesMap.REPORT_KEY: TrainingHistory()})

    @property
    def report_history(self) -> TrainingHistory:
        return self.histories[TrainingHistoriesMap.REPORT_KEY]

    def max(self, key=None):
        if self.curriculum is None:
            return self.histories[TrainingHistoriesMap.REPORT_KEY].max(key)
        else:
            return self.histories[self.curriculum.current_lesson.name].max(key)

    def concat(self, other):
        self.histories[TrainingHistoriesMap.REPORT_KEY].concat(other)
        if self.curriculum is not None:
            return self.histories[self.curriculum.current_lesson.name].concat(other)

    def append(self, key, value):
        self.histories[TrainingHistoriesMap.REPORT_KEY].append(key, value)
        if self.curriculum is not None:
            return self.histories[self.curriculum.current_lesson.name].append(key, value)

    @staticmethod
    def _set_default_plot_kwargs(kwargs: dict):
        kwargs.setdefault('fontsize', 16)
        kwargs.setdefault('linewidth', 3)
        kwargs.setdefault('figsize', (16, 12))
        kwargs.setdefault('dpi', 300)
        return kwargs

    def plot(self, save_path=None, show=False, lesson_idx: Optional[Union[int, str]] = None, **kwargs):
        kwargs = self._set_default_plot_kwargs(kwargs)
        if self.curriculum is None:
            assert lesson_idx is None, "lesson_idx must be None if curriculum is None"
            return self.plot_history(TrainingHistoriesMap.REPORT_KEY, save_path, show, **kwargs)
        if lesson_idx is None:
            self.plot_history(TrainingHistoriesMap.REPORT_KEY, save_path, show, **kwargs)
        else:
            self.plot_history(self.curriculum[lesson_idx].name, save_path, show, **kwargs)

    def plot_history(
            self,
            history_name: str,
            save_path=None,
            show=False,
            **kwargs
    ):
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        history = self.histories[history_name]
        if self.curriculum is not None and history_name != TrainingHistoriesMap.REPORT_KEY:
            lessons = [self.curriculum[history_name]]
            lessons_start_itr = [0]
        elif self.curriculum is not None and history_name == TrainingHistoriesMap.REPORT_KEY:
            lessons = self.curriculum.lessons
            lessons_lengths = {k: [len(self.histories[lesson.name][k]) for lesson in lessons] for k in history._container}
            lessons_start_itr = {k: np.cumsum(lessons_lengths[k]) for k in history.keys()}
        else:
            lessons = []
            lessons_start_itr = []

        kwargs = self._set_default_plot_kwargs(kwargs)
        loss_metrics = [k for k in history.keys() if 'loss' in k.lower()]
        rewards_metrics = [k for k in history.keys() if 'reward' in k.lower()]
        other_metrics = [k for k in history.keys() if k not in loss_metrics and k not in rewards_metrics]
        n_metrics = 2 + len(other_metrics)
        n_cols = int(np.sqrt(n_metrics))
        n_rows = int(n_metrics / n_cols)
        fig, axes = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=kwargs["figsize"], sharex='all')
        if axes.ndim == 1:
            axes = np.expand_dims(axes, axis=-1)
        for row_i in range(n_rows):
            for col_i in range(n_cols):
                ax = axes[row_i, col_i]
                ravel_index = row_i * n_cols + col_i
                if ravel_index == 0:
                    for k in loss_metrics:
                        ax.plot(history[k], label=k, linewidth=kwargs['linewidth'])
                    ax.set_ylabel("Loss [-]", fontsize=kwargs["fontsize"])
                    ax.legend(fontsize=kwargs["fontsize"])
                elif ravel_index == 1:
                    for k in rewards_metrics:
                        ax.plot(history[k], label=k, linewidth=kwargs['linewidth'])
                        for lesson_idx, lesson in enumerate(lessons):
                            if lesson.completion_criteria.measure == k:
                                ax.plot(
                                    lesson.completion_criteria.threshold*np.ones(len(history[k])), 'k--',
                                    label=f"{k} threshold", linewidth=kwargs['linewidth']
                                )
                            if history_name == TrainingHistoriesMap.REPORT_KEY and lesson.is_completed:
                                ax.axvline(
                                    lessons_start_itr[k][lesson_idx], ymin=np.min(history[k]), ymax=np.max(history[k]),
                                    color='r', linestyle='--', linewidth=kwargs['linewidth'], label=f"lesson start"
                                )
                    ax.set_ylabel("Rewards [-]", fontsize=kwargs["fontsize"])
                    ax.legend(fontsize=kwargs["fontsize"])
                else:
                    k = other_metrics[ravel_index - 1]
                    ax.plot(history[k], label=k, linewidth=kwargs['linewidth'])
                    ax.legend(fontsize=kwargs["fontsize"])
                if row_i == n_rows - 1:
                    ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
                legend_without_duplicate_labels_(ax)
        if save_path is not None:
            plt.savefig(save_path, dpi=kwargs["dpi"])
        if show:
            plt.show()
        plt.close(fig)


def space_to_spec(space: gym.spaces.Space) -> Dict[str, gym.spaces.Space]:
    spec = {}
    if hasattr(space, 'spaces'):
        for k, v in space.spaces.items():
            spec[k] = space_to_spec(v)
    else:
        spec[str(type(space).__name__)] = space
    return spec


def space_to_continuous_shape(
        space: gym.spaces.Space,
        flatten_spaces=False
) -> Union[Tuple[int, ...], Dict[str, Tuple[int, ...]]]:
    if hasattr(space, 'spaces'):
        shapes = {}
        for k, v in space.spaces.items():
            shapes[k] = space_to_continuous_shape(v)
        return shapes
    else:
        if isinstance(space, (gym.spaces.Discrete, )):
            return (space.n, )
        else:
            shape = space.shape
            if flatten_spaces:
                shape = (np.prod(shape), )
            return shape


def obs_sequence_to_batch(
        obs: Sequence[Union[np.ndarray, torch.Tensor, Dict[str, Union[np.ndarray, torch.Tensor]]]]
) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
    """
    Convert a sequence of observations to a batch of observations.

    :param obs: The sequence of observations.
    :type obs: Sequence[Union[np.ndarray, torch.Tensor, Dict[str, Union[np.ndarray, torch.Tensor]]]]

    :return: The batch of observations.
    :rtype: Union[torch.Tensor, Dict[str, torch.Tensor]]
    """
    obs_as_tensor = [to_tensor(obs_i) for obs_i in obs]
    if isinstance(obs_as_tensor[0], dict):
        return {k: torch.stack([o[k] for o in obs_as_tensor]) for k in obs[0].keys()}
    else:
        return torch.stack(obs_as_tensor)


def obs_batch_to_sequence(
        obs: Union[torch.Tensor, Dict[str, torch.Tensor]],
        as_numpy: bool = False
) -> Sequence[Union[np.ndarray, torch.Tensor, Dict[str, Union[np.ndarray, torch.Tensor]]]]:
    """
    Convert a batch of observations to a sequence of observations.

    :param obs: The batch of observations.
    :type obs: Union[torch.Tensor, Dict[str, torch.Tensor]]
    :param as_numpy: Whether to convert the observations to numpy arrays.
    :type as_numpy: bool

    :return: The sequence of observations.
    :rtype: Sequence[Union[np.ndarray, torch.Tensor, Dict[str, Union[np.ndarray, torch.Tensor]]]]
    """
    if as_numpy:
        obs = to_numpy(obs)
    if isinstance(obs, dict):
        return [{k: obs[k][i] for k in obs.keys()} for i in range(obs[list(obs.keys())[0]].shape[0])]
    else:
        return [obs[i] for i in range(obs.shape[0])]


class Linear(BaseNeuronsLayer):
    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        super().__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=False,
            device=device,
            **kwargs
        )
        self.bias_weights = None
        self.activation = self._init_activation(self.kwargs["activation"])

    def _set_default_kwargs(self):
        self.kwargs.setdefault("use_bias", True)
        self.kwargs.setdefault("activation", "identity")

    def _init_activation(self, activation: Union[torch.nn.Module, str]):
        """
        Initialise the activation function.
        :param activation: Activation function.
        :type activation: Union[torch.nn.Module, str]
        """
        str_to_activation = {
            "identity": torch.nn.Identity(),
            "relu"    : torch.nn.ReLU(),
            "tanh"    : torch.nn.Tanh(),
            "sigmoid" : torch.nn.Sigmoid(),
            "softmax" : torch.nn.Softmax(dim=-1),
        }
        if isinstance(activation, str):
            activation = activation.lower()
            assert activation in str_to_activation.keys(), f"Activation {activation} is not implemented."
            self.activation = str_to_activation[activation]
        else:
            self.activation = activation
        return self.activation

    def build(self) -> 'Linear':
        if self.kwargs["use_bias"]:
            self.bias_weights = torch.nn.Parameter(
                torch.empty((int(self.output_size),), device=self.device),
                requires_grad=self.requires_grad,
            )
        else:
            self.bias_weights = torch.zeros((int(self.output_size),), dtype=torch.float32, device=self.device)
        super().build()
        self.initialize_weights_()
        return self

    def initialize_weights_(self):
        super().initialize_weights_()
        torch.nn.init.kaiming_uniform_(self._forward_weights.data, a=np.sqrt(5))
        if "bias_weights" in self.kwargs:
            self.bias_weights.data = to_tensor(self.kwargs["bias_weights"]).to(self.device)
        else:
            # torch.nn.init.constant_(self.bias_weights, 0.0)
            bound = 1 / np.sqrt(int(self.input_size))
            torch.nn.init.uniform_(self.bias_weights, -bound, bound)

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        kwargs.setdefault("n_hh", 0)
        return super().create_empty_state(batch_size=batch_size, **kwargs)

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        # assert inputs.ndim == 2
        # batch_size, nb_features = inputs.shape
        # x = torch.functional.F.linear(inputs, self.forward_weights.T, self.bias_weights)
        x = torch.matmul(inputs, self.forward_weights) + self.bias_weights
        return self.activation(x)


def env_batch_step(
        env: gym.Env,
        actions: Any
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Step the environment in batch mode.

    :param env: The environment.
    :type env: gym.Env
    :param actions: The actions to take.
    :type actions: Any

    :return: The batch of observations, rewards, dones, truncated and infos.
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    # actions_as_numpy = to_numpy(actions).reshape(-1).tolist()  # TODO: to numpy without changing the dtype
    if isinstance(actions, dict) and len(actions) == 1:
        actions = list(actions.values())[0]
    actions_as_numpy = actions
    if isinstance(env, gym.vector.VectorEnv):
        observations, rewards, dones, truncateds, info = env.step(actions_as_numpy)
        infos = [info for _ in range(env.num_envs)]
    else:
        actions_as_single = actions_as_numpy[0] if actions_as_numpy.ndim > 0 else actions_as_numpy
        observation, reward, done, truncated, info = env.step(actions_as_single)
        observations = batch_dict_of_items(observation)
        rewards = np.array([reward])
        dones = np.array([done])
        truncateds = np.array([truncated])
        infos = np.array([info])
    return observations, rewards, dones, truncateds, infos



def env_batch_reset(env: gym.Env) -> Tuple[np.ndarray, np.ndarray]:
    """
    Reset the environment in batch mode.

    :param env: The environment.
    :type env: gym.Env

    :return: The batch of observations.
    :rtype: np.ndarray
    """
    if isinstance(env, gym.vector.VectorEnv):
        observations, infos = env.reset()
    else:
        observation, info = env.reset()
        observations = batch_dict_of_items(observation)
        infos = np.array([info])
    return observations, infos


def env_batch_render(env: gym.Env, **kwargs) -> List[Any]:
    """
    Render the environment in batch mode.

    :param env: The environment.
    :type env: gym.Env
    """
    if isinstance(env, gym.vector.VectorEnv):
        rendering = env.render()
        if rendering is None:
            rendering = [None for _ in range(env.num_envs)]
    else:
        rendering = env.render()
        rendering = [rendering]
    return rendering


def get_single_observation_space(env: gym.Env) -> gym.spaces.Space:
    """
    Return the observation space of a single environment.

    :param env: The environment.
    :type env: gym.Env

    :return: The observation space.
    :rtype: gym.spaces.Space
    """
    if isinstance(env, gym.vector.VectorEnv):
        return env.single_observation_space
    else:
        return env.observation_space


def get_single_action_space(env: gym.Env) -> gym.spaces.Space:
    """
    Return the action space of a single environment.

    :param env: The environment.
    :type env: gym.Env

    :return: The action space.
    :rtype: gym.spaces.Space
    """
    if isinstance(env, gym.vector.VectorEnv):
        return env.single_action_space
    else:
        return env.action_space


def sample_action_space(action_space: gym.spaces.Space, re_format: str = "raw"):
    """
    Sample an action from the action space.

    :param action_space: The action space.
    :type action_space: gym.spaces.Space
    :param re_format: The format to return the action in.
    :type re_format: str

    :return: The sampled action.
    :rtype: Any
    """
    if isinstance(action_space, (gym.spaces.Discrete, gym.spaces.MultiDiscrete, gym.spaces.Box)):
        action = action_space.sample()
    else:
        raise NotImplementedError(f"Action space {action_space} is not implemented.")
    if re_format == "raw":
        return action
    elif re_format == "one_hot":
        if isinstance(action_space, gym.spaces.Discrete):
            return np.eye(action_space.n)[action]
        elif isinstance(action_space, gym.spaces.MultiDiscrete):
            return np.stack([np.eye(n)[a] for n, a in zip(action_space.nvec, action)])
        else:
            raise NotImplementedError(f"Action space {action_space} is not implemented.")
    else:
        raise NotImplementedError(f"Format {re_format} is not implemented.")


def discounted_cumulative_sums(x, discount, axis=-1, **kwargs):
    # Discounted cumulative sums of vectors for computing rewards-to-go and advantage estimates
    from scipy.signal import lfilter
    conv = lfilter([1], [1, float(-discount)], x[::-1], axis=axis)[::-1]
    return conv


def batch_numpy_actions(actions, env: Optional[gym.Env] = None):
    actions_as_numpy = to_numpy(actions)
    if isinstance(actions_as_numpy, np.ndarray):
        dim = len(actions_as_numpy.shape)
        if dim == 0:
            return np.expand_dims(actions_as_numpy, axis=0)
    elif isinstance(actions_as_numpy, dict):
        return {k: batch_numpy_actions(v) for k, v in actions_as_numpy.items()}
    else:
        raise NotImplementedError(f"Type {type(actions_as_numpy)} is not implemented.")
    if env is not None:
        actions_as_numpy = format_numpy_actions(actions_as_numpy, env)
    return actions_as_numpy


def format_numpy_actions(actions, env: gym.Env):
    actions_as_numpy = to_numpy(actions)
    entry_is_dict = isinstance(actions_as_numpy, dict)
    spec = space_to_spec(env.action_space)
    if not entry_is_dict:
        assert len(spec) == 1, f"Expected only one action space, but got {len(spec)}."
        actions_as_numpy = {k: actions_as_numpy for k in spec}
    for k, space in spec.items():
        if isinstance(space, (gym.spaces.Discrete, gym.spaces.MultiDiscrete)):
            actions_as_numpy[k] = actions_as_numpy[k].astype(np.int64)
        elif isinstance(space, gym.spaces.Box):
            actions_as_numpy[k] = actions_as_numpy[k].astype(np.float32)
        else:
            pass
    if not entry_is_dict:
        actions_as_numpy = actions_as_numpy[list(actions_as_numpy.keys())[0]]
    return actions_as_numpy


class TrajectoryRenderer:
    def __init__(
            self,
            trajectory: Trajectory,
            env: Optional[gym.Env] = None,
            **kwargs
    ):
        self.trajectory = trajectory
        self.env = env
        if self.check_simulate_is_needed():
            if self.env is None:
                raise ValueError(
                    "If the experiences in the trajectory do not have others['render'], an environment must be provided."
                )
            self.simulate()

    def check_simulate_is_needed(self):
        is_needed = not all("render" in x.others for x in self.trajectory)
        return is_needed

    def simulate(self):
        warnings.warn(
            "This method is deprecated. Its highly recommended to set the the attribute others['render'] "
            "when generating trajectories. If you are using RLAcademy.generate_trajectories, you can set "
            "the argument render=True to do so easily."
        )
        for x in self.trajectory:
            self.env.unwrapped.state = x.obs
            x.others["render"] = self.env.render()

    def render(self, **kwargs) -> Tuple[plt.Figure, plt.Axes, mpl_animation.FuncAnimation]:
        filename = kwargs.get("filename", None)
        file_extension = kwargs.get("file_extension", "gif")
        writer = kwargs.get("writer", "ffmpeg")
        fps = kwargs.get("fps", 30)
        time_interval = 1 / fps

        fig = kwargs.get("fig", None)
        ax = kwargs.get("ax", None)
        if fig is None or ax is None:
            fig, ax = plt.subplots()
        env_name = self.env.unwrapped.spec.id
        title = f"Trajectory on {env_name}.\nCumulative reward: {self.trajectory.cumulative_reward:.2f}."
        title = kwargs.get("title", title)
        fig.suptitle(title, fontsize=kwargs.get("title_font_size", 16))
        ax.axis("off")
        im = ax.imshow(self.trajectory[0].others["render"])

        def _animation(i):
            im.set_array(self.trajectory[i].others["render"])
            return im,

        anim = mpl_animation.FuncAnimation(
            fig, _animation, frames=len(self.trajectory), interval=time_interval, blit=True
        )
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            if file_extension is None:
                if '.' in filename:
                    file_extension = filename.split('.')[-1]
                else:
                    file_extension = 'gif'
            assert file_extension in ["mp4", "gif"], "The extension of the file must be mp4 or gif."
            if filename.endswith(file_extension):
                filename = ''.join(filename.split('.')[:-1])
            anim.save(f"{filename}.{file_extension}", writer=writer, fps=fps)
        if kwargs.get("show", True):
            plt.show()
        return fig, ax, anim

    def to_file(self, file_path: str, fps: int = 30, **kwargs):
        import imageio

        if "." in file_path:
            ext = os.path.splitext(file_path)[-1]
        else:
            ext = kwargs.get("ext", ".mp4")
        if not file_path.endswith(ext):
            file_path += ext
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with imageio.get_writer(file_path, fps=fps) as writer:
            for x in self.trajectory:
                writer.append_data(x.others["render"])
        return file_path

    def to_mp4(self, file_path: str, fps: int = 30, **kwargs):
        return self.to_file(file_path, fps=fps, ext="mp4", **kwargs)

    def to_gif(self, file_path: str, fps: int = 30, **kwargs):
        return self.to_file(file_path, fps=fps, ext="gif", **kwargs)


def continuous_actions_distribution(
        actions: Union[Dict, torch.Tensor, np.ndarray],
        covariance: Optional[Union[Dict, torch.Tensor, np.ndarray]] = None
) -> Union[Dict, torch.distributions.Distribution]:
    """
    Creates a continuous action distribution from the actions and the covariance.

    :param actions: The actions.
    :type actions: Union[Dict, torch.Tensor, np.ndarray]
    :param covariance: The covariance of the actions. If None, a diagonal covariance is assumed using the variance of
        the given actions.
    :type covariance: Optional[Union[Dict, torch.Tensor, np.ndarray]]
    :return: The action distribution.
    :rtype: Union[Dict, torch.distributions.Distribution]
    """
    if isinstance(actions, dict):
        dist = {
            k: continuous_actions_distribution(actions[k], covariance[k] if covariance is not None else None)
            for k in actions
        }
    else:
        actions = to_tensor(actions)
        if covariance is None:
            std = torch.std(actions.view(-1, actions.shape[-1]), dim=0)
            covariance = torch.diag(std**2)
        else:
            covariance = to_tensor(covariance)
        dist = torch.distributions.MultivariateNormal(actions, covariance.to(actions.device))
    return dist


def batch_dict_of_items(x: Any) -> Any:
    if isinstance(x, dict):
        return {k: batch_dict_of_items(v) for k, v in x.items()}
    else:
        return np.array([x])


def get_item_from_batch(x: Any, i: int) -> Any:
    if isinstance(x, dict):
        return {k: get_item_from_batch(v, i) for k, v in x.items()}
    else:
        return x[i]



from collections import defaultdict
from typing import Optional, Sequence, Union, Dict, Callable, Tuple

import torch
import torch.nn.functional as F

from .learning_algorithm import LearningAlgorithm
from ..learning_algorithms.tbptt import TBPTT
from ..transforms.base import ToDevice
from ..utils import list_insert_replace_at, ConnectivityConvention, unpack_out_hh, format_pred_batch
from ..utils.autograd import compute_jacobian, filter_parameters, recursive_detach, recursive_detach_


class RLS(TBPTT):
    r"""
    Apply the recursive least squares algorithm to the given model. Different strategies are available to update the
    parameters of the model. The strategy is defined by the :attr:`strategy` attribute of the class. The following
    strategies are available:

        - `inputs`: The parameters are updated using the inputs of the model.
        - `outputs`: The parameters are updated using the outputs of the model. This one is inspired by the work of
            Perich and al. :cite:t:`perich_inferring_2021` with the CURBD algorithm.
        - `grad`: The parameters are updated using the gradients of the model. This one is inspired by the work of
            Zhang and al. :cite:t:`zhang_revisiting_2021`.
        - `jacobian`: The parameters are updated using the Jacobian of the model. This one is inspired by the work of
            Al-Batah and al. :cite:t:`al-batah_modified_2010`.
        - `scaled_jacobian`: The parameters are updated using the scaled Jacobian of the model.

    .. note::
        The `inputs` and `outputs` strategies are limited to an optimization of only one parameter. The others
        strategies can be used with multiple parameters. Unfortunately, those strategies do not work as expected
        at the moment. If you want to help with the development of those strategies, please open an issue on
        GitHub.

    """
    CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
    CHECKPOINT_P_STATES_DICT_KEY: str = "P_list"

    @staticmethod
    def curbd_step_method(
            inv_corr: torch.Tensor,
            post_activation: torch.Tensor,
            error: torch.Tensor,
            connectivity_convention: ConnectivityConvention = ConnectivityConvention.ItoJ,
            **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        phi = torch.mean(post_activation.view(-1, post_activation.shape[-1]), dim=0).view(-1, 1)  # [f_out, 1]
        k = torch.matmul(inv_corr, phi)  # [f_out, f_out] @ [f_out, 1] -> [f_out, 1]
        rPr = torch.matmul(phi.T, k)  # [1, f_out] @ [f_out, 1] -> [1]
        c = 1.0 / (1.0 + rPr)  # [1]
        delta_inv_corr = c * torch.matmul(k, k.T)  # [f_out, 1] @ [1, f_out] -> [f_out, f_out]
        if connectivity_convention == ConnectivityConvention.ItoJ:
            delta_weights = c * torch.outer(k.view(-1), error.view(-1))  # [f_out, 1] @ [1, f_out] -> [N_in, N_out]
        elif connectivity_convention == ConnectivityConvention.JtoI:
            delta_weights = c * torch.outer(error.view(-1), k.view(-1))  # [f_out, 1] @ [1, f_out] -> [N_in, N_out]
        else:
            raise ValueError(f"Invalid connectivity convention: {connectivity_convention}")
        return delta_weights, delta_inv_corr

    def __init__(
            self,
            *,
            params: Optional[Sequence[torch.nn.Parameter]] = None,
            layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
            criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
            backward_time_steps: Optional[int] = None,
            is_recurrent: bool = False,
            **kwargs
    ):
        """
        Constructor for RLS class.

        :param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
        :type params: Optional[Sequence[torch.nn.Parameter]]
        :param layers: The layers to optimize. If not None the parameters of the layers will be added to the
            parameters to optimize.
        :type layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]]
        :param criterion: The criterion to use. If not provided, torch.nn.MSELoss is used.
        :type criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]]
        :param backward_time_steps: The frequency of parameter optimisation. If None, the number of
            time steps of the data will be used.
        :type backward_time_steps: Optional[int]
        :param is_recurrent: If True, the model is recurrent. If False, the model is not recurrent.
        :type is_recurrent: bool
        :param kwargs: The keyword arguments to pass to the BaseCallback.

        :keyword bool save_state: Whether to save the state of the optimizer. Defaults to True.
        :keyword bool load_state: Whether to load the state of the optimizer. Defaults to True.
        """
        kwargs.setdefault("auto_backward_time_steps_ratio", 0)
        kwargs.setdefault("save_state", True)
        kwargs.setdefault("load_state", True)
        kwargs.setdefault("weight_decay", 0.0)
        super().__init__(
            params=params,
            layers=layers,
            criterion=criterion,
            backward_time_steps=backward_time_steps,
            optimizer=None,
            optim_time_steps=None,
            **kwargs
        )

        # RLS attributes
        self.P_list = None
        self.delta = kwargs.get("delta", 1.0)
        self.Lambda = kwargs.get("Lambda", 1.0)
        self.kappa = kwargs.get("kappa", 1.0)
        self._device = kwargs.get("device", None)
        self.to_cpu_transform = ToDevice(device=torch.device("cpu"))
        self.to_device_transform = None
        self._other_dims_as_batch = kwargs.get("other_dims_as_batch", False)
        self._is_recurrent = is_recurrent
        self.strategy = kwargs.get("strategy", "inputs").lower()
        self.strategy_to_mth = {
            "inputs": self.inputs_mth_step,
            "outputs": self.outputs_mth_step,
            "grad": self.grad_mth_step,
            "jacobian": self.jacobian_mth_step,
            "scaled_jacobian": self.scaled_jacobian_mth_step,
        }
        self.kwargs = kwargs
        self._asserts()
        self._last_layers_buffer = defaultdict(list)

    def _asserts(self):
        assert 0.0 < self.Lambda <= 1.0, "Lambda must be between 0 and 1"
        assert self.strategy in self.strategy_to_mth, f"Strategy must be one of {list(self.strategy_to_mth.keys())}"

    def __repr__(self):
        repr_str = f"{self.name}: ("
        repr_str += f"priority={self.priority}, "
        repr_str += f"save_state={self.save_state}, "
        repr_str += f"load_state={self.load_state}, "
        repr_str += f"strategy={self.strategy}, "
        repr_str += f"delta={self.delta}, "
        repr_str += f"Lambda={self.Lambda}, "
        repr_str += f"kappa={self.kappa}"
        repr_str += f")"
        return repr_str

    def initialize_P_list(self, m=None):
        self.P_list = [
            self.delta * torch.eye(param.numel() if m is None else m, dtype=torch.float32, device=torch.device("cpu"))
            for param in self.params
        ]

    def _maybe_update_time_steps(self):
        if self._auto_set_backward_time_steps:
            self.backward_time_steps = max(1, int(self._auto_backward_time_steps_ratio * self._data_n_time_steps))

    def _decorate_forward(self, forward, layer_name: str):
        def _forward(*args, **kwargs):
            out = forward(*args, **kwargs)
            t = kwargs.get("t", None)
            if t is None:
                return out
            out_tensor, hh = unpack_out_hh(out)
            list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
            if len(self._layers_buffer[layer_name]) == self.backward_time_steps:
                self._backward_at_t(t, self.backward_time_steps, layer_name)
                if self.strategy in ["grad", "jacobian", "scaled_jacobian"]:
                    out = recursive_detach(out)
            return out
        return _forward

    def _output_hook(self, module, args, kwargs, output) -> None:
        t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
        if t is None:
            return

        layer_name = module.name
        out_tensor, hh = unpack_out_hh(output)
        list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
        if len(self._layers_buffer[layer_name]) == self.backward_time_steps:
            self._backward_at_t(t, self.backward_time_steps, layer_name)
            if self.strategy in ["grad", "jacobian", "scaled_jacobian"]:
                output = recursive_detach_(output)

    def _backward_at_t(self, t: int, backward_t: int, layer_name: str):
        if self._last_layers_buffer[layer_name]:
            x_batch = self._get_x_batch_from_buffer(layer_name)
        else:
            x_batch = self._get_x_batch_slice_from_trainer(0, backward_t, layer_name)
        y_batch = self._get_y_batch_slice_from_trainer((t + 1) - backward_t, t + 1, layer_name)
        pred_batch = self._get_pred_batch_from_buffer(layer_name)
        self.optimization_step(x_batch, pred_batch, y_batch)
        self._last_layers_buffer[layer_name] = self._layers_buffer[layer_name].copy()
        self._layers_buffer[layer_name].clear()

    def _get_x_batch_slice_from_trainer(self, t_first: int, t_last: int, layer_name: str = None):
        x_batch = self.trainer.current_training_state.x_batch
        if isinstance(x_batch, dict):
            if layer_name is None:
                x_batch = {
                    key: val[:, t_first:t_last]
                    for key, val in x_batch.items()
                }
            else:
                x_batch = x_batch[layer_name][:, t_first:t_last]
        else:
            x_batch = x_batch[:, t_first:t_last]
        return x_batch.clone()

    def _get_x_batch_from_buffer(self, layer_name: str):
        pred_batch = torch.stack(self._last_layers_buffer[layer_name], dim=1)
        return pred_batch

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        if self.save_state:
            state = checkpoint.get(self.name, {})
            self.P_list = state.get(self.CHECKPOINT_P_STATES_DICT_KEY, None)

    def get_checkpoint_state(self, trainer, **kwargs) -> object:
        if self.save_state:
            return {
                self.CHECKPOINT_P_STATES_DICT_KEY: self.P_list,
            }
        return None

    def _try_put_on_device(self, trainer):
        try:
            self.P_list = [self.to_device_transform(p) for p in self.P_list]
        except Exception as e:
            trainer.model = self.to_cpu_transform(trainer.model)
            self.P_list = [self.to_device_transform(p) for p in self.P_list]

    def _put_on_cpu(self):
        self.P_list = [self.to_cpu_transform(p) for p in self.P_list]

    def start(self, trainer, **kwargs):
        super().start(trainer, **kwargs)
        if self._device is None:
            self._device = trainer.model.device
        self.to_device_transform = ToDevice(device=self._device)
        self.params = filter_parameters(self.params, requires_grad=True)

    def on_batch_begin(self, trainer, **kwargs):
        LearningAlgorithm.on_batch_begin(self, trainer, **kwargs)
        self.trainer = trainer
        if self._is_recurrent:
            self._data_n_time_steps = self._get_data_time_steps_from_y_batch(
                trainer.current_training_state.y_batch, trainer.current_training_state.x_batch
            )
            self._maybe_update_time_steps()
            self.decorate_forwards()

    def on_batch_end(self, trainer, **kwargs):
        super().on_batch_end(trainer)
        self.undecorate_forwards()
        self._layers_buffer.clear()

    def on_optimization_begin(self, trainer, **kwargs):
        x_batch = trainer.current_training_state.x_batch
        y_batch = trainer.current_training_state.y_batch
        pred_batch = format_pred_batch(trainer.current_training_state.pred_batch, y_batch)

        if self._is_recurrent:
            for layer_name in self._layers_buffer:
                backward_t = len(self._layers_buffer[layer_name])
                if backward_t > 0:
                    self._backward_at_t(self._data_n_time_steps - 1, backward_t, layer_name)
        else:
            self.optimization_step(x_batch, pred_batch, y_batch)

        trainer.update_state_(batch_loss=self.apply_criterion(pred_batch, y_batch).detach_())

    def optimization_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
        if self.strategy not in self.strategy_to_mth:
            raise ValueError(f"Invalid strategy: {self.strategy}")
        return self.strategy_to_mth[self.strategy](x_batch, pred_batch, y_batch)

    def scaled_jacobian_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
        """
        This method is inspired by the work of Al-Batah and al. :cite:t:`al-batah_modified_2010`. Unfortunately, this
        method does not seem to work with the current implementation.

        TODO: Make it work.

        x.shape = [B, f_in]
        y.shape = [B, f_out]
        error.shape = [B, f_out]
        P.shape = [f_out, f_out]
        theta.shape = [ell, 1]

        epsilon = mean[B](error[B, f_out]) -> [1, f_out]
        phi = mean[B](y[B, f_out]) [1, f_out]
        psi = jacobian[theta](phi[1, f_out]]) -> [f_out, ell]

        K = P[f_out, f_out] @ psi[f_out, ell] -> [f_out, ell]
        h = 1 / (labda[1] + kappa[1] * psi.T[ell, f_out] @ K[f_out, ell]) -> [ell, ell]
        grad = (K[f_out, ell] @ h[ell, ell]).T[ell, f_out] @ epsilon.T[f_out, 1] -> [1, ell]
        P = labda[1] * P[f_out, f_out] - kappa[1] * K[f_out, ell] @ h[ell, ell] @ K[f_out, ell].T -> [f_out, f_out]

        In this case f_in must be equal to N_in.

        :param x_batch: inputs of the layer
        :param pred_batch: outputs of the layer
        :param y_batch: targets of the layer

        """
        if getattr(self.trainer, "model", None) is None:
            model_device = pred_batch.device
        else:
            model_device = self.trainer.model.device
        assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
        assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
        assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
        self.optimizer.zero_grad()

        x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
        pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
        y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
        error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]

        if self.P_list is None:
            self.initialize_P_list(m=pred_batch_view.shape[-1])
        self.P_list = self.to_device_transform(self.P_list)
        self.params = self.to_device_transform(self.params)

        epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
        phi = pred_batch_view.mean(dim=0).view(1, -1)  # [1, f_out]
        psi_list = compute_jacobian(params=self.params, y=phi.view(-1), strategy="slow")  # [f_out, ell]
        K_list = [torch.matmul(P, psi) for P, psi in zip(self.P_list, psi_list)]  # [f_out, f_out] @ [f_out, ell] -> [f_out, ell]
        h_list = [torch.linalg.pinv(self.Lambda + self.kappa * torch.matmul(psi.T, K)) for psi, K in zip(psi_list, K_list)]  # [ell, f_out] @ [f_out, ell] -> [ell, ell]

        for p, K, h in zip(self.params, K_list, h_list):
            p.grad = torch.matmul(torch.matmul(K, h).T, epsilon.T).view(p.shape).clone()  # ([f_out, ell] @ [ell, ell]).T @ [f_out, 1] -> [ell, 1]

        self.optimizer.step()
        self.P_list = [
            self.Lambda * P - self.kappa * torch.matmul(torch.matmul(K, h), K.T)
            for P, h, K in zip(self.P_list, h_list, K_list)
        ]  # [f_out, f_out] - [f_out, ell] @ [ell, ell] @ [ell, f_out] -> [f_out, f_out]

        self._put_on_cpu()
        if getattr(self.trainer, "model", None) is not None:
            self.trainer.model.to(model_device, non_blocking=True)

    def jacobian_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
        """
        This method is inspired by the work of Al-Batah and al. :cite:t:`al-batah_modified_2010`. Unfortunately, this
        method does not seem to work with the current implementation.

        TODO: Make it work.

        x.shape = [B, f_in]
        y.shape = [B, f_out]
        error.shape = [B, f_out]
        P.shape = [f_out, f_out]
        theta.shape = [ell, 1]

        epsilon = mean[B](error[B, f_out]) -> [1, f_out]
        phi = mean[B](y[B, f_out]) [1, f_out]
        psi = jacobian[theta](phi[1, f_out]]) -> [f_out, L]

        K = P[f_out, f_out] @ psi[f_out, L] -> [f_out, L]
        grad = epsilon[1, f_out] @ K[f_out, L] -> [L, 1]
        P = labda[1] * P[f_out, f_out] - kappa[1] * K[f_out, L] @ K[f_out, L].T -> [f_out, f_out]

        In this case f_in must be equal to N_in.

        :param x_batch: inputs of the layer
        :param pred_batch: outputs of the layer
        :param y_batch: targets of the layer

        """
        if getattr(self.trainer, "model", None) is None:
            model_device = pred_batch.device
        else:
            model_device = self.trainer.model.device
        assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
        assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
        assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
        self.optimizer.zero_grad()

        x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
        pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
        y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
        error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]

        if self.P_list is None:
            self.initialize_P_list(m=pred_batch_view.shape[-1])
        self.P_list = self.to_device_transform(self.P_list)
        self.params = self.to_device_transform(self.params)

        epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
        phi = pred_batch_view.mean(dim=0).view(1, -1)  # [1, f_out]
        psi_list = compute_jacobian(params=self.params, y=phi.view(-1), strategy="slow")  # [f_out, L]
        K_list = [torch.matmul(P, psi) for P, psi in zip(self.P_list, psi_list)]  # [f_out, f_out] @ [f_out, ell] -> [f_out, L]

        for p, K in zip(self.params, K_list):
            p.grad = torch.matmul(K.T, epsilon.T).view(p.shape).clone()  # [L, f_out] @ [f_out, 1] -> [L, 1]

        self.optimizer.step()
        self.P_list = [
            self.Lambda * P - self.kappa * torch.matmul(K, K.T)
            for P, K in zip(self.P_list, K_list)
        ]  # [f_out, f_out] - [f_out, L] @ [L, f_out] -> [f_out, f_out]

        self._put_on_cpu()
        if getattr(self.trainer, "model", None) is not None:
            self.trainer.model.to(model_device, non_blocking=True)

    def grad_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
        """
        This method is inspired by the work of Zhang and al. :cite:t:`zhang_revisiting_2021`. Unfortunately, this
        method does not seem to work with the current implementation.

        TODO: Make it work.

        x.shape = [B, f_in]
        y.shape = [B, f_out]
        error.shape = [B, f_out]
        P.shape = [f_in, f_in]

        epsilon = mean[B](error[B, f_out]) -> [1, f_out]
        phi = mean[B](x[B, f_in]) [1, f_in]

        K = P[f_in, f_in] @ phi.T[f_in, 1] -> [f_in, 1]
        h = 1 / (labda[1] + kappa[1] * phi[1, f_in] @ K[f_in, 1]) -> [1]
        grad = h[1] * P[f_in, f_in] @ grad[N_in, N_out] -> [N_in, N_out]
        P = labda[1] * P[f_in, f_in] - h[1] * kappa[1] * K[f_in, 1] @ K.T[1, f_in] -> [f_in, f_in]

        In this case f_in must be equal to N_in.

        :param x_batch: inputs of the layer
        :param pred_batch: outputs of the layer
        :param y_batch: targets of the layer

        """
        if getattr(self.trainer, "model", None) is None:
            model_device = pred_batch.device
        else:
            model_device = self.trainer.model.device
        assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
        assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
        assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
        self.optimizer.zero_grad()

        mse_loss = F.mse_loss(pred_batch, y_batch)
        if mse_loss.grad_fn is None:
            # TODO: add a check if it is always the case and if so, warn the user.
            return
        mse_loss.backward()

        x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
        pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
        y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
        error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]

        if self.P_list is None:
            self.initialize_P_list(m=x_batch_view.shape[-1])
            for p in self.params:
                # making sur that f_in = N_in.
                if p.shape[0] != x_batch_view.shape[-1]:
                    raise ValueError(
                        f"For inputs of shape [B, f_in], the first dimension of the parameters must be f_in, "
                        f"got {p.shape[0]} instead of {x_batch_view.shape[-1]}."
                    )
        self.P_list = self.to_device_transform(self.P_list)
        self.params = self.to_device_transform(self.params)

        epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
        phi = x_batch_view.mean(dim=0).view(1, -1).detach().clone()  # [1, f_in]
        K_list = [torch.matmul(P, phi.T) for P in self.P_list]  # [f_in, f_in] @ [f_in, 1] -> [f_in, 1]
        h_list = [1.0 / (self.Lambda + self.kappa * torch.matmul(phi, K)).item() for K in K_list]  # [1, f_in] @ [f_in, 1] -> [1]

        for p, P, h in zip(self.params, self.P_list, h_list):
            p.grad = h * torch.matmul(P, p.grad)  # [f_in, f_in] @ [N_in, N_out] -> [N_in, N_out]

        self.optimizer.step()
        self.P_list = [
            self.Lambda * P - h * self.kappa * torch.matmul(K, K.T)
            for P, h, K in zip(self.P_list, h_list, K_list)
        ]  # [f_in, 1] @ [1, f_in] -> [f_in, f_in]

        self._put_on_cpu()
        if getattr(self.trainer, "model", None) is not None:
            self.trainer.model.to(model_device, non_blocking=True)

    def inputs_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
        """


        x.shape = [B, f_in]
        y.shape = [B, f_out]
        error.shape = [B, f_out]
        P.shape = [f_in, f_in]

        epsilon = mean[B](error[B, f_out]) -> [1, f_out]
        phi = mean[B](x[B, f_in]) [1, f_in]

        K = P[f_in, f_in] @ phi.T[f_in, 1] -> [f_in, 1]
        h = 1 / (labda[1] + kappa[1] * phi[1, f_in] @ K[f_in, 1]) -> [1]
        P = labda[1] * P[f_in, f_in] - h[1] * kappa[1] * K[f_in, 1] @ K.T[1, f_in] -> [f_in, f_in]
        grad = h[1] * K[f_in, 1] @ epsilon[1, f_out] -> [N_in, N_out]

        In this case [N_in, N_out] must be equal to [f_in, f_out].

        :param x_batch: inputs of the layer
        :param pred_batch: outputs of the layer
        :param y_batch: targets of the layer

        """
        if getattr(self.trainer, "model", None) is None:
            model_device = pred_batch.device
        else:
            model_device = self.trainer.model.device
        assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
        assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
        assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
        self.optimizer.zero_grad()

        x_batch_view = x_batch.view(-1, x_batch.shape[-1]).detach()  # [B, f_in]
        pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1]).detach()  # [B, f_out]
        y_batch_view = y_batch.view(-1, y_batch.shape[-1]).detach()  # [B, f_out]
        error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]

        if self.P_list is None:
            self.initialize_P_list(m=x_batch_view.shape[-1])
            for p in self.params:
                # making sur that f_in = N_in.
                if p.shape[0] != x_batch_view.shape[-1]:
                    raise ValueError(
                        f"For inputs of shape [B, f_in], the first dimension of the parameters must be f_in, "
                        f"got {p.shape[0]} instead of {x_batch_view.shape[-1]}."
                    )
                # making sure that f_out = N_out.
                if p.shape[1] != y_batch_view.shape[-1]:
                    raise ValueError(
                        f"For targets of shape [B, f_out], the second dimension of the parameters must be f_out, "
                        f"got {p.shape[1]} instead of {y_batch_view.shape[-1]}."
                    )
        self.P_list = self.to_device_transform(self.P_list)
        self.params = self.to_device_transform(self.params)

        epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
        phi = x_batch_view.mean(dim=0).view(1, -1).detach().clone()  # [1, f_in]
        K_list = [torch.matmul(P, phi.T) for P in self.P_list]  # [f_in, f_in] @ [f_in, 1] -> [f_in, 1]
        h_list = [1.0 / (self.Lambda + self.kappa * torch.matmul(phi, K)).item() for K in K_list]  # [1, f_in] @ [f_in, 1] -> [1]

        for p, K, h in zip(self.params, K_list, h_list):
            p.grad = h * torch.outer(K.view(-1), epsilon.view(-1))  # [f_in, 1] @ [1, f_out] -> [N_in, N_out]

        self.optimizer.step()
        self.P_list = [
            self.Lambda * P - h * self.kappa * torch.matmul(K, K.T)
            for P, h, K in zip(self.P_list, h_list, K_list)
        ]  # [f_in, 1] @ [1, f_in] -> [f_in, f_in]

        self._put_on_cpu()
        if getattr(self.trainer, "model", None) is not None:
            self.trainer.model.to(model_device, non_blocking=True)

    def outputs_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
        """
        This method is inspired by the work of Perich and al. :cite:t:`perich_inferring_2021` with
        the CURBD algorithm.

        x.shape = [B, f_in]
        y.shape = [B, f_out]
        error.shape = [B, f_out]
        epsilon = mean[B](error[B, f_out]) -> [1, f_out]
        phi = mean[B](y[B, f_out]) [1, f_out]

        P.shape = [f_out, f_out]
        K = P[f_out, f_out] @ phi.T[f_out, 1] -> [f_out, 1]
        h = 1 / (labda[1] + kappa[1] * phi[1, f_out] @ K[f_out, 1]) -> [1]
        P = labda[1] * P[f_out, f_out] - h[1] * kappa[1] * K[f_out, 1] @ K.T[1, f_out] -> [f_out, f_out]
        grad = h[1] * K[f_out, 1] @ epsilon[1, f_out] -> [N_in, N_out]

        In this case [N_in, N_out] must be equal to [f_out, f_out].

        :param x_batch: inputs of the layer
        :param pred_batch: outputs of the layer
        :param y_batch: targets of the layer

        """
        if getattr(self.trainer, "model", None) is None:
            model_device = pred_batch.device
        else:
            model_device = self.trainer.model.device
        assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
        assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
        assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
        self.optimizer.zero_grad()

        x_batch_view = x_batch.view(-1, x_batch.shape[-1]).detach()  # [B, f_in]
        pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1]).detach()  # [B, f_out]
        y_batch_view = y_batch.view(-1, y_batch.shape[-1]).detach()  # [B, f_out]
        error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]

        if self.P_list is None:
            self.initialize_P_list(m=pred_batch_view.shape[-1])
            for p in self.params:
                # making sur that f_out = N_in.
                if p.shape[0] != pred_batch_view.shape[-1]:
                    raise ValueError(
                        f"For inputs of shape [B, f_in], the first dimension of the parameters must be f_in, "
                        f"got {p.shape[0]} instead of {x_batch_view.shape[-1]}."
                    )
                # making sure that f_out = N_out.
                if p.shape[1] != pred_batch_view.shape[-1]:
                    raise ValueError(
                        f"For targets of shape [B, f_out], the second dimension of the parameters must be f_out, "
                        f"got {p.shape[1]} instead of {y_batch_view.shape[-1]}."
                    )
        assert len(self.params) == 1, "This method support the optimization of only one recurrent matrix."
        assert len(self.P_list) == len(self.params), "The number of parameters and P matrices must be equal."
        self.P_list = self.to_device_transform(self.P_list)
        self.params = self.to_device_transform(self.params)

        epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]

        cubd_objs = [
            self.curbd_step_method(P, pred_batch_view, epsilon)
            for P in self.P_list
        ]
        for i, (p, (delta_weights, delta_inv_corr)) in enumerate(zip(self.params, cubd_objs)):
            p.grad = delta_weights
            self.P_list[i] = self.Lambda * self.P_list[i] - self.kappa * delta_inv_corr

        # phi = pred_batch_view.mean(dim=0).view(1, -1).detach().clone()  # [1, f_out]
        # K_list = [torch.matmul(P, phi.T) for P in self.P_list]  # [f_out, f_out] @ [f_out, 1] -> [f_out, 1]
        # h_list = [1.0 / (self.Lambda + self.kappa * torch.matmul(phi, K)).item() for K in K_list]  # [1, f_out] @ [f_out, 1] -> [1]
        #
        # for p, K, h in zip(self.params, K_list, h_list):
        # 	p.grad = h * torch.outer(K.view(-1), epsilon.view(-1))  # [f_out, 1] @ [1, f_out] -> [N_in, N_out]

        self.optimizer.step()
        # self.P_list = [
        # 	self.Lambda * P - h * self.kappa * torch.matmul(K, K.T)
        # 	for P, h, K in zip(self.P_list, h_list, K_list)
        # ]  # [f_in, 1] @ [1, f_in] -> [f_in, f_in]

        self._put_on_cpu()
        if getattr(self.trainer, "model", None) is not None:
            self.trainer.model.to(model_device, non_blocking=True)




from typing import Any, Type, Union, Optional

import numpy as np
import torch

from . import ConstantValuesTransform, to_tensor
from ..modules.layers import LIFLayer, SpyLIFLayer, ALIFLayer


class SpikesEncoder(torch.nn.Module):
    def __init__(
            self,
            n_steps: int,
            n_units: int,
            spikes_layer_type: Type[Union[LIFLayer, SpyLIFLayer, ALIFLayer]],
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        """
        Constructor for the SpikesEncoder class.

        :param n_steps: Number of steps to encode.
        :param n_units: Number of units in the encoder.
        :param spikes_layer_type: Dynamic of the spikes layer.
        :param dt: Time step of the simulation.
        :param device: Device to use for the encoder.
        :param kwargs: Keyword arguments for the spikes layer.

        :keyword kwargs:
            * <forward_weights>: Forward weights of the spikes layer. Default: eye(n_units).
            * <use_recurrent_connection>: Whether to use a recurrent connection. Default: False.
            * <name>: Name of the spikes layer. Default: "encoder".

        """
        super().__init__()
        self.n_steps = n_steps
        self.const_transform = ConstantValuesTransform(n_steps, batch_wise=True)
        self.spikes_layer_type = spikes_layer_type
        self.dt = dt
        kwargs.setdefault('forward_weights', torch.eye(n_units))
        kwargs.setdefault('use_recurrent_connection', False)
        kwargs.setdefault('name', 'encoder')
        kwargs.setdefault('freeze_weights', False)

        assert "dt" not in kwargs, \
            "dt cannot be specified since it must be the given dt"
        assert "device" not in kwargs, \
            "device cannot be specified since it must be the given device"

        self.spikes_layer = self.spikes_layer_type(
            n_units, n_units,
            dt=dt,
            device=device,
            **kwargs,
        )
        self.spikes_layer.build()

    def to(self, device):
        self.spikes_layer.device = device
        return super().to(device)

    def forward(self, x: Any):
        x = to_tensor(x).to(self.spikes_layer.device)
        if x.ndim < 3:
            x = x.unsqueeze(1)
        x = self.const_transform(x)
        x_spikes, hh = [], None
        for t in range(self.n_steps):
            spikes, hh = self.spikes_layer(x[:, t], hh)
            x_spikes.append(spikes)
        # x_spikes = torch.squeeze(torch.stack(x_spikes, dim=1))
        return torch.stack(x_spikes, dim=1)

    def get_and_reset_regularization_loss(self) -> torch.Tensor:
        """
        Get the regularization loss as a sum of all the regularization losses of the layers. Then reset the
        regularization losses.

        :return: the regularization loss.
        """
        regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.spikes_layer.device)
        if hasattr(self.spikes_layer, "get_and_reset_regularization_loss") and callable(
                self.spikes_layer.get_and_reset_regularization_loss
        ):
            regularization_loss += self.spikes_layer.get_and_reset_regularization_loss()
        return regularization_loss


class LIFEncoder(SpikesEncoder):
    def __init__(
            self,
            n_steps: int,
            n_units: int,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            spikes_layer_kwargs: Optional[dict] = None,
    ):
        if spikes_layer_kwargs is None:
            spikes_layer_kwargs = {}
        super().__init__(
            n_steps, n_units, LIFLayer, dt, device, **spikes_layer_kwargs,
        )


class SpyLIFEncoder(SpikesEncoder):
    def __init__(
            self,
            n_steps: int,
            n_units: int,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            spikes_layer_kwargs: Optional[dict] = None,
    ):
        if spikes_layer_kwargs is None:
            spikes_layer_kwargs = {}
        super().__init__(
            n_steps, n_units, SpyLIFLayer, dt, device, **spikes_layer_kwargs,
        )


class ALIFEncoder(SpikesEncoder):
    def __init__(
            self,
            n_steps: int,
            n_units: int,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            spikes_layer_kwargs: Optional[dict] = None,
    ):
        if spikes_layer_kwargs is None:
            spikes_layer_kwargs = {}
        super().__init__(
            n_steps, n_units, ALIFLayer, dt, device, **spikes_layer_kwargs,
        )


